import os
from unittest import mock
import pytest
import responses
from datetime import datetime

from github3.repos.repo import Repository
from github3.pulls import ShortPullRequest
from github3.exceptions import ConnectionError
from github3.session import AppInstallationTokenAuth

from cumulusci.core import github
from cumulusci.core.exceptions import GithubException
from cumulusci.tasks.release_notes.tests.utils import MockUtil
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.core.github import (
    create_gist,
    get_github_api,
    validate_service,
    create_pull_request,
    markdown_link_to_pr,
    is_pull_request_merged,
    get_github_api_for_repo,
    is_label_on_pull_request,
    get_pull_requests_by_head,
    add_labels_to_pull_request,
    get_pull_requests_by_commit,
    get_pull_requests_with_base_branch,
)


class TestGithub(GithubApiTestMixin):
    @classmethod
    def teardown_method(cls):
        # clear cached repo -> installation mapping
        github.INSTALLATIONS.clear()

    @pytest.fixture
    def mock_util(self):
        return MockUtil("TestOwner", "TestRepo")

    @pytest.fixture
    def repo(self, gh_api):
        repo_json = self._get_expected_repo("TestOwner", "TestRepo")
        return Repository(repo_json, gh_api)

    def test_github_api_retries(self, mock_http_response):
        gh = get_github_api("TestUser", "TestPass")
        adapter = gh.session.get_adapter("http://")

        assert 0.3 == adapter.max_retries.backoff_factor
        assert 502 in adapter.max_retries.status_forcelist

        with mock.patch(
            "urllib3.connectionpool.HTTPConnectionPool._make_request"
        ) as _make_request:
            _make_request.side_effect = [
                mock_http_response(status=503),
                mock_http_response(status=200),
            ]

            gh.octocat("meow")
            assert 2 == _make_request.call_count

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            json={
                "id": 1,
                "access_tokens_url": "",
                "account": "",
                "app_id": "",
                "created_at": "",
                "events": "",
                "html_url": "",
                "permissions": "",
                "repositories_url": "",
                "repository_selection": "",
                "single_file_name": "",
                "target_id": "",
                "target_type": "",
                "updated_at": "",
            },
        )
        responses.add(
            "POST",
            "https://api.github.com/app/installations/1/access_tokens",
            status=201,
            json={"token": "ITOKEN", "expires_at": datetime.now().isoformat()},
        )

        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            gh = get_github_api_for_repo(None, "TestOwner", "TestRepo")
            assert isinstance(gh.session.auth, AppInstallationTokenAuth)

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo__not_installed(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            status=404,
        )
        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            with pytest.raises(GithubException):
                get_github_api_for_repo(None, "TestOwner", "TestRepo")

    @responses.activate
    def test_validate_service(self):
        responses.add("GET", "https://api.github.com/rate_limit", status=401)
        with pytest.raises(GithubException):
            validate_service({"username": "BOGUS", "password": "BOGUS"})

    @responses.activate
    def test_get_pull_requests_by_head(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(
            pulls=self._get_expected_pull_requests(1),
            head=repo.owner.login + ":" + "some-other-branch",
        )
        pull_requests = get_pull_requests_by_head(repo, "some-other-branch")
        assert 1 == len(pull_requests)

        # ConnectionError present when we reachout with
        # a branch name (url parameter) that we aren't expecting
        with pytest.raises(ConnectionError):
            get_pull_requests_by_head(repo, "does-not-exist")

    @responses.activate
    def test_get_pull_requests_by_head__no_pulls(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls()
        pull_requests = get_pull_requests_by_head(repo, "test_branch")
        assert pull_requests == []

        pull_requests = get_pull_requests_by_head(repo, "master")
        assert pull_requests is None

    @responses.activate
    def test_get_pull_request_by_head__multiple_pulls(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(pulls=self._get_expected_pull_requests(2))
        pull_requests = get_pull_requests_by_head(repo, "test_branch")
        assert 2 == len(pull_requests)

    @responses.activate
    def test_get_pull_requests_with_base_branch(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(base="master", head="TestOwner:some-branch")
        pull_requests = get_pull_requests_with_base_branch(
            repo, "master", head="some-branch"
        )
        assert 0 == len(pull_requests)

        responses.reset()
        mock_util.mock_pulls(pulls=self._get_expected_pull_requests(3), base="master")
        pull_requests = get_pull_requests_with_base_branch(repo, "master")
        assert 3 == len(pull_requests)

    @responses.activate
    def test_create_pull_request(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.mock_pulls(
            method=responses.POST,
            pulls=self._get_expected_pull_request(1, 1, "Test Body"),
        )
        pull_request = create_pull_request(repo, "test-branch")
        assert pull_request is not None
        assert pull_request.body == "Test Body"

    @responses.activate
    def test_is_label_on_pr(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.add_issue_response(self._get_expected_issue(1))
        mock_util.add_issue_response(self._get_expected_issue(2))
        mock_util.mock_issue_labels(1, responses.GET, ["Octocat", "bogus 1", "bogus 2"])
        mock_util.mock_issue_labels(2, responses.GET, ["bogus 1", "bogus 2"])

        pull_request = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        pull_request.number = 1
        assert is_label_on_pull_request(repo, pull_request, "Octocat")
        pull_request.number = 2
        assert not is_label_on_pull_request(repo, pull_request, "Octocat")

    @responses.activate
    def test_add_labels_to_pull_request(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.add_issue_response(self._get_expected_issue(1))
        mock_util.mock_issue_labels(1, method=responses.POST)
        pull_request = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        pull_request.number = 1

        add_labels_to_pull_request(repo, pull_request, "first", "second", "third")
        body = responses.calls[-1].request.body
        assert "first" in body
        assert "second" in body
        assert "third" in body

    @responses.activate
    def test_get_pull_request_by_commit(self, mock_util, repo, gh_api):
        self.init_github()
        commit_sha = "asdf1234asdf1234"
        mock_util.mock_pull_request_by_commit_sha(commit_sha)
        pull_requests = get_pull_requests_by_commit(gh_api, repo, commit_sha)
        assert len(pull_requests) == 1

    def test_is_pull_request_merged(self, gh_api):
        self.init_github()

        merged_pull_request = ShortPullRequest(
            self._get_expected_pull_request(1, 1), gh_api
        )
        merged_pull_request.merged_at = "DateTimeStr"

        unmerged_pull_request = ShortPullRequest(
            self._get_expected_pull_request(1, 1), gh_api
        )
        unmerged_pull_request.merged_at = None

        assert is_pull_request_merged(merged_pull_request)
        assert not is_pull_request_merged(unmerged_pull_request)

    def test_markdown_link_to_pr(self, gh_api):
        self.init_github()
        pr = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        actual_link = markdown_link_to_pr(pr)
        expected_link = f"{pr.title} [[PR{pr.number}]({pr.html_url})]"

        assert expected_link == actual_link

    @responses.activate
    def test_create_gist(self, gh_api, mock_util):
        self.init_github()

        description = "Test Gist Creation"
        filename = "error_output.txt"
        content = "Hello there gist!"
        files = {filename: content}

        self.mock_gist(description, files)
        gist = create_gist(gh_api, description, files)

        expected_url = f"https://gist.github.com/{gist.id}"
        assert expected_url == gist.html_url

    def mock_gist(self, description, files):
        responses.add(
            method=responses.POST,
            url=f"https://api.github.com/gists",
            json=self._get_expected_gist(description, files),
            status=201,
        )



from cumulusci.salesforce_api.metadata import ApiRetrieveInstalledPackages
from cumulusci.tasks.salesforce import BaseSalesforceMetadataApiTask


class GetInstalledPackages(BaseSalesforceMetadataApiTask):
    api_class = ApiRetrieveInstalledPackages
    name = "GetInstalledPackages"



import unittest

from unittest import mock
import responses

from cumulusci.tasks import bulkdata
from cumulusci.tasks.bulkdata.tests.test_bulkdata import _make_task
from cumulusci.core.exceptions import TaskOptionsError

BULK_DELETE_QUERY_RESULT = b"Id\n003000000000001".splitlines()
BULK_DELETE_RESPONSE = b'<root xmlns="http://ns"><id>4</id></root>'
BULK_BATCH_RESPONSE = '<root xmlns="http://ns"><batch><state>{}</state></batch></root>'


@mock.patch("cumulusci.tasks.bulkdata.delete.time.sleep", mock.Mock())
class TestDeleteData(unittest.TestCase):
    def _configure_mocks(self, query_job, query_batch, delete_job):
        api = mock.Mock()
        api.endpoint = "http://api"
        api.jobNS = "http://ns"
        api.create_query_job.return_value = query_job
        api.query.return_value = query_batch
        api.is_batch_done.side_effect = [False, True, False, True]
        api.get_all_results_for_query_batch.return_value = [BULK_DELETE_QUERY_RESULT]
        api.create_job.return_value = delete_job
        api.headers.return_value = {}
        responses.add(
            method="POST",
            url="http://api/job/3/batch",
            body=BULK_DELETE_RESPONSE,
            status=200,
        )
        api.job_status.return_value = {
            "numberBatchesCompleted": 1,
            "numberBatchesTotal": 1,
        }
        responses.add(
            method="GET",
            url="http://api/job/3/batch",
            body=BULK_BATCH_RESPONSE.format("InProgress"),
            status=200,
        )
        responses.add(
            method="GET",
            url="http://api/job/3/batch",
            body=BULK_BATCH_RESPONSE.format("Completed"),
            status=200,
        )
        return api

    @responses.activate
    def test_run(self):
        query_job = "1"
        query_batch = "2"
        delete_job = "3"

        api = self._configure_mocks(query_job, query_batch, delete_job)
        task = _make_task(bulkdata.DeleteData, {"options": {"objects": "Contact"}})

        def _init_class():
            task.bulk = api

        task._init_class = _init_class
        task()

        api.create_query_job.assert_called_once_with("Contact", contentType="CSV")
        api.query.assert_called_once_with(query_job, "SELECT Id FROM Contact")
        api.is_batch_done.assert_has_calls(
            [mock.call(query_batch, query_job), mock.call(query_batch, query_job)]
        )
        api.create_job.assert_called_once_with("Contact", "delete")
        api.close_job.assert_has_calls([mock.call(query_job), mock.call(delete_job)])

    @responses.activate
    def test_run_with_where(self):
        query_job = "1"
        query_batch = "2"
        delete_job = "3"

        api = self._configure_mocks(query_job, query_batch, delete_job)
        task = _make_task(
            bulkdata.DeleteData,
            {"options": {"objects": "Contact", "where": "city='Goshen'"}},
        )

        def _init_class():
            task.bulk = api

        task._init_class = _init_class
        task()

        api.create_query_job.assert_called_once_with("Contact", contentType="CSV")
        api.query.assert_called_once_with(
            query_job, "SELECT Id FROM Contact WHERE city='Goshen'"
        )
        api.is_batch_done.assert_has_calls(
            [mock.call(query_batch, query_job), mock.call(query_batch, query_job)]
        )
        api.create_job.assert_called_once_with("Contact", "delete")
        api.close_job.assert_has_calls([mock.call(query_job), mock.call(delete_job)])

    def test_create_job__no_records(self):
        task = _make_task(bulkdata.DeleteData, {"options": {"objects": "Contact"}})
        task._query_salesforce_for_records_to_delete = mock.Mock(return_value=[])
        task.logger = mock.Mock()
        task._create_job("Contact")
        task.logger.info.assert_called_with(
            "  No Contact objects found, skipping delete"
        )

    def test_parse_job_state(self):
        task = _make_task(bulkdata.DeleteData, {"options": {"objects": "Contact"}})
        api = mock.Mock()
        api.jobNS = "http://ns"
        task.bulk = api
        self.assertEqual(
            ("InProgress", None),
            task._parse_job_state(
                '<root xmlns="http://ns">'
                "  <batch><state>InProgress</state></batch>"
                "  <batch><state>Failed</state><stateMessage>test</stateMessage></batch>"
                "  <batch><state>Completed</state></batch>"
                "</root>"
            ),
        )
        self.assertEqual(
            ("Failed", ["test"]),
            task._parse_job_state(
                '<root xmlns="http://ns">'
                "  <batch><state>Failed</state><stateMessage>test</stateMessage></batch>"
                "  <batch><state>Completed</state></batch>"
                "</root>"
            ),
        )
        self.assertEqual(
            ("Completed", None),
            task._parse_job_state(
                '<root xmlns="http://ns">'
                "  <batch><state>Completed</state></batch>"
                "  <batch><state>Completed</state></batch>"
                "</root>"
            ),
        )
        self.assertEqual(
            ("Aborted", None),
            task._parse_job_state(
                '<root xmlns="http://ns">'
                "  <batch><state>Not Processed</state></batch>"
                "  <batch><state>Completed</state></batch>"
                "</root>"
            ),
        )
        self.assertEqual(
            ("CompletedWithFailures", ["Failures detected: 200"]),
            task._parse_job_state(
                '<root xmlns="http://ns">'
                "  <batch><state>Completed</state></batch>"
                "  <numberRecordsFailed>200</numberRecordsFailed>"
                "</root>"
            ),
        )

    @responses.activate
    def test_upload_batches__error(self):
        task = _make_task(bulkdata.DeleteData, {"options": {"objects": "Contact"}})
        api = mock.Mock()
        api.endpoint = "http://api"
        api.headers.return_value = {}
        api.raise_error.side_effect = Exception

        def _init_class():
            task.bulk = api

        task._init_class = _init_class
        responses.add(responses.POST, "http://api/job/1/batch", body=b"", status=500)
        with self.assertRaises(Exception):
            list(task._upload_batches("1", [{"Id": "1"}]))

    def test_validate_options(self):
        with self.assertRaises(TaskOptionsError):
            _make_task(bulkdata.DeleteData, {"options": {"objects": ""}})

        with self.assertRaises(TaskOptionsError):
            _make_task(
                bulkdata.DeleteData, {"options": {"objects": "a,b", "where": "x='y'"}}
            )



from cumulusci.core.exceptions import ApexCompilationException
from cumulusci.core.exceptions import ApexException
from cumulusci.core.exceptions import SalesforceException
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils import in_directory


class AnonymousApexTask(BaseSalesforceApiTask):
    """ Executes anonymous apex from a file or string."""

    task_docs = """
    Use the `apex` option to run a string of anonymous Apex.
    Use the `path` option to run anonymous Apex from a file.
    Or use both to concatenate the string to the file contents.
    """

    task_options = {
        "path": {"description": "The path to an Apex file to run.", "required": False},
        "apex": {
            "description": "A string of Apex to run (after the file, if specified).",
            "required": False,
        },
        "managed": {
            "description": (
                "If True, will insert the project's namespace prefix.  "
                "Defaults to False or no namespace."
            ),
            "required": False,
        },
        "namespaced": {
            "description": (
                "If True, the tokens %%%NAMESPACED_RT%%% and %%%namespaced%%% "
                "will get replaced with the namespace prefix for Record Types."
            ),
            "required": False,
        },
        "param1": {
            "description": (
                "Parameter to pass to the Apex. Use as %%%PARAM_1%%% in the Apex code."
                "Defaults to an empty value."
            ),
            "required": False,
        },
        "param2": {
            "description": (
                "Parameter to pass to the Apex. Use as %%%PARAM_2%%% in the Apex code."
                "Defaults to an empty value."
            ),
            "required": False,
        },
    }

    def _validate_options(self):
        super(AnonymousApexTask, self)._validate_options()

        if not self.options.get("path") and not self.options.get("apex"):
            raise TaskOptionsError(
                "You must specify either the `path` or `apex` option."
            )

    def _run_task(self):
        apex = ""
        apex_path = self.options.get("path")
        if apex_path:
            if not in_directory(apex_path, self.project_config.repo_root):
                raise TaskOptionsError(
                    "Please specify a path inside your project repository. "
                    "You specified: {}".format(apex_path)
                )
            self.logger.info("Executing anonymous Apex from {}".format(apex_path))
            try:
                with open(apex_path, "r") as f:
                    apex = f.read()
            except IOError:
                raise TaskOptionsError(
                    "Could not find or read file: {}".format(apex_path)
                )
        else:
            self.logger.info("Executing anonymous Apex")

        apex_string = self.options.get("apex")
        if apex_string:
            apex = apex + "\n" + apex_string

        apex = self._prepare_apex(apex)
        result = self.tooling._call_salesforce(
            method="GET",
            url="{}executeAnonymous".format(self.tooling.base_url),
            params={"anonymousBody": apex},
        )
        self._check_result(result)

        self.logger.info("Anonymous Apex Success")

    def _prepare_apex(self, apex):
        # Process namespace tokens
        managed = self.options.get("managed") or False
        namespaced = self.options.get("namespaced") or False
        namespace = self.project_config.project__package__namespace
        namespace_prefix = ""
        record_type_prefix = ""
        if managed or namespaced:
            namespace_prefix = namespace + "__"
        if namespaced:
            record_type_prefix = namespace + "."
        apex = apex.replace("%%%NAMESPACE%%%", namespace_prefix)
        apex = apex.replace("%%%NAMESPACED_ORG%%%", namespace_prefix)
        apex = apex.replace("%%%NAMESPACED_RT%%%", record_type_prefix)

        # Process optional parameter token replacement
        param1 = self.options.get("param1") or ""
        apex = apex.replace("%%%PARAM_1%%%", param1)
        param2 = self.options.get("param2") or ""
        apex = apex.replace("%%%PARAM_2%%%", param2)

        return apex

    def _check_result(self, result):
        # anon_results is an ExecuteAnonymous Result
        # https://developer.salesforce.com/docs/atlas.en-us.apexcode.meta/apexcode/sforce_api_calls_executeanonymous_result.htm
        anon_results = result.json()

        # A result of `None` (body == "null") with a 200 status code
        # means that a gack occurred.
        if anon_results is None:
            raise SalesforceException(
                "Anonymous Apex returned the result `null`. "
                "This often indicates a gack occurred."
            )
        if not anon_results["compiled"]:
            raise ApexCompilationException(
                anon_results["line"], anon_results["compileProblem"]
            )
        if not anon_results["success"]:
            raise ApexException(
                anon_results["exceptionMessage"], anon_results["exceptionStackTrace"]
            )



lex_locators = {
    "app_launcher": {
        "app_link": "//div[@class='slds-section slds-is-open salesforceIdentityAppLauncherDesktopInternal']//section[@id='cards']//a[@class='appTileTitle' and text()='{}']",
        "button": "//nav[contains(@class,'appLauncher')]//button[//div[contains(@class,'slds-icon-waffle')]]",
        "current_app": "//div[contains(@class,'navLeft')]//span[contains(@class,'appName')]/span[text()='{}']",
        "tab_link": "css: a.app-launcher-link[title='{}']",
    },
    "desktop_rendered": "css: div.desktop.container.oneOne.oneAppLayoutHost[data-aura-rendered-by]",
    "loading_box": "css: div.auraLoadingBox.oneLoadingBox",
    "modal": {
        "button": "//div[contains(@class,'uiModal')]//div[contains(@class,'modal-footer')]//button[.//span[text()='{}']]",
        "close": "css: button.slds-modal__close",
        "error_messages": "css: div.pageLevelErrors ul.errorsList li",
        "fade_in": "css: div.slds-fade-in-open",
        "has_error": "css: div.pageLevelErrors",
        "is_open": "css: div.uiModal div.panel.slds-modal",
    },
    "object": {
        "button": "css: div.windowViewMode-normal ul.forceActionsContainer.oneActionsRibbon a[title='{}']",
        "field": "//div[contains(@class, 'uiInput')][.//label[contains(@class, 'uiLabel')][.//span[text()='{}']]]//*[self::input or self::textarea]",
        "field_lookup_link": "//a[@role='option'][.//div[@title='{}']]",
        "field_lookup_value": "//div[contains(@class, 'uiInput')][.//label[contains(@class, 'uiLabel')][.//span[text()='{}']]]//span[contains(@class,'pillText')]",
        "record_type_option": "//div[contains(@class, 'changeRecordTypeOptionRightColumn')]//span[text()='{}']",
    },
    "record": {
        "header": {
            "field": "//li[contains(@class, 'slds-page-header__detail-block')][.//span[contains(@class, 'slds-form-element__label')][@title='{}']]",
            "field_value": "//li[contains(@class, 'slds-page-header__detail-block')][.//span[contains(@class, 'slds-form-element__label')][@title='{}']]//div[contains(@class, 'slds-form-element__static')]/span[text()]",
            "field_value_link": "//li[contains(@class, 'slds-page-header__detail-block')][.//span[contains(@class, 'slds-form-element__label')][@title='{}']]//div[contains(@class, 'slds-form-element__static')]//a",
            "field_value_checked": "//li[contains(@class, 'slds-page-header__detail-block')][.//span[contains(@class, 'slds-form-element__label')][@title='{}']]//span[contains(@class, 'uiOutputCheckbox')]//img[@alt='True']",
            "field_value_unchecked": "//li[contains(@class, 'slds-page-header__detail-block')][.//span[contains(@class, 'slds-form-element__label')][@title='{}']]//span[contains(@class, 'uiOutputCheckbox')]//img[@alt='False']",
        },
        "related": {
            "card": "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img][.//span[@title='{}']]",
            "button": "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img][.//span[@title='{}']]//a[@title='{}']",
            "count": "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img]//span[@title='{}']/following-sibling::span",
            "link": "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img][.//span[@title='{}']]//table[contains(@class,'forceRecordLayout')]/tbody/tr[.//th/div/a[contains(@class,'textUnderline')]][.//td//a[text()='{}']]/th//a",
            "popup_trigger": "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img][.//span[@title='{}']]//tr[.//a[text()='{}']]//div[contains(@class, 'forceVirtualAction')]//a",
        },
    },
    "popup": {
        "link": "//div[contains(@class, 'uiPopupTarget')][contains(@class, 'visible')]//a[@title='{}']"
    },
    "spinner": "css: div.slds-spinner",
    "tabs": {"tab": ""},
    "actions": "css: ul.oneActionsRibbon",
    "body": "//div[contains(@class, 'slds-template__container')]/*",
}



import base64
import os
import pickle

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC

from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.exceptions import ConfigError, KeychainKeyNotFound
from cumulusci.core.keychain import BaseProjectKeychain

BS = 16
backend = default_backend()


def pad(s):
    return s + (BS - len(s) % BS) * chr(BS - len(s) % BS).encode("ascii")


class BaseEncryptedProjectKeychain(BaseProjectKeychain):
    """ Base class for building project keychains that use AES encryption for securing stored org credentials """

    encrypted = True

    def _get_connected_app(self):
        if self.app:
            return self._decrypt_config(ConnectedAppOAuthConfig, self.app)

    def _get_service(self, name):
        return self._decrypt_config(ServiceConfig, self.services[name])

    def _set_service(self, service, service_config, project):
        encrypted = self._encrypt_config(service_config)
        self._set_encrypted_service(service, encrypted, project)

    def _set_encrypted_service(self, service, encrypted, project):
        self.services[service] = encrypted

    def _set_org(self, org_config, global_org):
        encrypted = self._encrypt_config(org_config)
        self._set_encrypted_org(org_config.name, encrypted, global_org)

    def _set_encrypted_org(self, name, encrypted, global_org):
        self.orgs[name] = encrypted

    def _get_org(self, name):
        return self._decrypt_config(OrgConfig, self.orgs[name], extra=[name])

    def _get_cipher(self, iv=None):
        key = self.key
        if not isinstance(key, bytes):
            key = key.encode()
        if iv is None:
            iv = os.urandom(16)
        cipher = Cipher(AES(key), CBC(iv), backend=backend)
        return cipher, iv

    def _encrypt_config(self, config):
        pickled = pickle.dumps(config.config, protocol=2)
        pickled = pad(pickled)
        cipher, iv = self._get_cipher()
        return base64.b64encode(iv + cipher.encryptor().update(pickled))

    def _decrypt_config(self, config_class, encrypted_config, extra=None):
        if not encrypted_config:
            if extra:
                return config_class(None, *extra)
            else:
                return config_class()
        encrypted_config = base64.b64decode(encrypted_config)
        iv = encrypted_config[:16]
        cipher, iv = self._get_cipher(iv)
        pickled = cipher.decryptor().update(encrypted_config[16:])
        unpickled = pickle.loads(pickled, encoding="bytes")
        # Convert bytes created in Python 2
        config_dict = {}
        for k, v in unpickled.items():
            if isinstance(k, bytes):
                k = k.decode("utf-8")
            if isinstance(v, bytes):
                v = v.decode("utf-8")
            config_dict[k] = v
        args = [config_dict]
        if extra:
            args += extra
        return self._construct_config(config_class, args)

    def _construct_config(self, config_class, args):
        if args[0].get("scratch"):
            config_class = ScratchOrgConfig

        return config_class(*args)

    def _validate_key(self):
        if not self.key:
            raise KeychainKeyNotFound("The keychain key was not found.")
        if len(self.key) != 16:
            raise ConfigError("The keychain key must be 16 characters long.")



import json
import requests
from datetime import datetime
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.exceptions import SalesforceException


class CreateCommunity(BaseSalesforceApiTask):
    api_version = "46.0"
    task_docs = """
    Create a Salesforce Community via the Connect API.
    Specify the `template` "VF Template" for Visualforce Tabs community,
    or the name for a specific desired template
    """
    task_options = {
        "template": {
            "description": "Name of the template for the community.",
            "required": True,
        },
        "name": {"description": "Name of the community.", "required": True},
        "description": {
            "description": "Description of the community.",
            "required": False,
        },
        "url_path_prefix": {
            "description": "URL prefix for the community.",
            "required": False,
        },
        "timeout": {
            "description": "Time to wait, in seconds, for the community to be created",
            "default": 120,
        },
    }

    def _init_options(self, kwargs):
        super(CreateCommunity, self)._init_options(kwargs)
        self.options["timeout"] = int(self.options.get("timeout", 120))

    def _run_task(self):
        self.logger.info('Creating community "{}"'.format(self.options["name"]))

        # Before we can create a Community, we have to click through the "New Community"
        # button in the All Communities setup page. (This does some unknown behind-the-scenes setup).
        # Let's simulate that without actually using a browser.
        self.logger.info("Preparing org for Communities")
        s = requests.Session()
        s.get(self.org_config.start_url).raise_for_status()
        r = s.get(
            "{}/sites/servlet.SitePrerequisiteServlet".format(
                self.org_config.instance_url
            )
        )
        if r.status_code != 200:
            raise SalesforceException("Unable to prepare org for Communities")

        payload = {
            "name": self.options["name"],
            "description": self.options.get("description") or "",
            "templateName": self.options["template"],
            "urlPathPrefix": self.options.get("url_path_prefix") or "",
        }

        self.logger.info("Sending request to create Community")
        self.sf.restful("connect/communities", method="POST", data=json.dumps(payload))

        # Wait for the community to be created
        self.time_start = datetime.now()
        self._poll()

    def _poll_action(self):
        elapsed = datetime.now() - self.time_start
        if elapsed.total_seconds() > self.options["timeout"]:
            raise SalesforceException(
                "Community creation not finished after {timeout} seconds".format(
                    **self.options
                )
            )

        community_list = self.sf.restful("connect/communities")["communities"]
        communities = {c["name"]: c for c in community_list}
        if self.options["name"] in communities:
            self.poll_complete = True
            self.logger.info(
                "Community {} created".format(communities[self.options["name"]]["id"])
            )



from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.exceptions import CumulusCIFailure


class MetadataApiError(CumulusCIFailure):
    def __init__(self, message, response):
        super(MetadataApiError, self).__init__(message)
        self.response = response


class MetadataComponentFailure(MetadataApiError):
    pass


class MetadataParseError(MetadataApiError):
    pass


class MissingOAuthError(CumulusCIException):
    pass


class MissingOrgCredentialsError(CumulusCIException):
    pass



from pathlib import Path
from unittest import mock

import pytest
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce import UpdateProfile

from .util import create_task

ADMIN_PROFILE_BEFORE = """<?xml version='1.0' encoding='utf-8'?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>npsp__Nonprofit_CRM</application>
        <default>true</default>
        <visible>false</visible>
    </applicationVisibilities>
    <classAccess>
        <apexClass>TestClass</apexClass>
        <enabled>false</enabled>
    </classAccess>
    <fieldPermissions>
        <field>Account.TestField__c</field>
        <editable>false</editable>
        <readable>false</readable>
    </fieldPermissions>
    <pageAccesses>
        <apexPage>TestPage</apexPage>
        <enabled>false</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Account.HH_Account</recordType>
        <default>false</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>false</visible>
    </recordTypeVisibilities>
    <tabVisibilities>
        <tab>NPSP_Settings</tab>
        <visibility>Hidden</visibility>
    </tabVisibilities>
</Profile>"""

ADMIN_PROFILE_EXPECTED = """<?xml version='1.0' encoding='utf-8'?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>npsp__Nonprofit_CRM</application>
        <default>true</default>
        <visible>true</visible>
    </applicationVisibilities>
    <classAccess>
        <apexClass>TestClass</apexClass>
        <enabled>true</enabled>
    </classAccess>
    <fieldPermissions>
        <field>Account.TestField__c</field>
        <editable>true</editable>
        <readable>true</readable>
    </fieldPermissions>
    <pageAccesses>
        <apexPage>TestPage</apexPage>
        <enabled>true</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>false</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Account.HH_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <tabVisibilities>
        <tab>NPSP_Settings</tab>
        <visibility>DefaultOn</visibility>
    </tabVisibilities>
</Profile>"""


def test_run_task():
    task = create_task(
        UpdateProfile,
        {
            "record_types": [
                {
                    "record_type": "Account.HH_Account",
                    "default": True,
                    "person_account_default": True,
                }
            ],
            "namespaced_org": True,
        },
    )

    def _retrieve_unpackaged():
        profiles_path = Path(task.retrieve_dir, "profiles")
        admin_profile_path = Path(profiles_path, "Admin.profile")
        profiles_path.mkdir()
        admin_profile_path.write_text(ADMIN_PROFILE_BEFORE)

    def _check_result():
        result_path = Path(task.retrieve_dir, "profiles", "Admin.profile")
        result = result_path.read_text()
        assert ADMIN_PROFILE_EXPECTED == result

    task._retrieve_unpackaged = _retrieve_unpackaged
    task._deploy_metadata = _check_result
    task()


def test_run_task__other_profile():
    task = create_task(
        UpdateProfile,
        {
            "record_types": [
                {
                    "record_type": "Account.HH_Account",
                    "default": True,
                    "person_account_default": True,
                }
            ],
            "namespaced_org": True,
            "profile_name": "Continuous Integration",
        },
    )

    def _retrieve_unpackaged():
        profiles_path = Path(task.retrieve_dir, "profiles")
        admin_profile_path = Path(profiles_path, "Continuous Integration.profile")
        profiles_path.mkdir()
        admin_profile_path.write_text(ADMIN_PROFILE_BEFORE)

    def _check_result():
        result_path = Path(
            task.retrieve_dir, "profiles", "Continuous Integration.profile"
        )
        result = result_path.read_text()
        assert ADMIN_PROFILE_EXPECTED == result

    task._retrieve_unpackaged = _retrieve_unpackaged
    task._deploy_metadata = _check_result
    task()


def test_run_task__record_type_not_found():
    task = create_task(
        UpdateProfile,
        {"record_types": [{"record_type": "DOESNT_EXIST"}], "namespaced_org": True},
    )

    def _retrieve_unpackaged():
        profiles_path = Path(task.retrieve_dir, "profiles")
        admin_profile_path = Path(profiles_path, "Admin.profile")
        profiles_path.mkdir()
        admin_profile_path.write_text(ADMIN_PROFILE_BEFORE)

    task._retrieve_unpackaged = _retrieve_unpackaged
    with pytest.raises(TaskOptionsError):
        task()


@mock.patch("cumulusci.salesforce_api.metadata.ApiRetrieveUnpackaged.__call__")
def test_retrieve_unpackaged(ApiRetrieveUnpackaged):
    task = create_task(UpdateProfile)
    task.retrieve_dir = "/tmp"
    task._retrieve_unpackaged()
    ApiRetrieveUnpackaged.assert_called_once()


def test_deploy_metadata(tmpdir):
    task = create_task(UpdateProfile)
    task.retrieve_dir = Path(tmpdir, "retrieve", "profiles")
    task.deploy_dir = Path(tmpdir, "deploy")
    task.retrieve_dir.mkdir(parents=True)
    task.deploy_dir.mkdir(parents=True)
    task.retrieve_dir = task.retrieve_dir.parent
    task.deploy_dir = task.deploy_dir.parent

    task._get_api = mock.Mock()
    task._deploy_metadata()
    task._get_api.assert_called_once()


def test_get_deploy_package_xml_content():
    task = create_task(UpdateProfile, {"profile_name": "Test"})

    assert "Test" in task._get_deploy_package_xml_content()

    task = create_task(UpdateProfile)

    assert "Admin" in task._get_deploy_package_xml_content()


def test_get_retrieve_package_xml_content():
    task = create_task(UpdateProfile, {"profile_name": "Test"})

    assert "Test" in task._get_retrieve_package_xml_content()

    task = create_task(UpdateProfile)

    assert "Admin" in task._get_retrieve_package_xml_content()



import os
import sys

from robot import run as robot_run
from robot.testdoc import testdoc

from cumulusci.core.exceptions import RobotTestFailure
from cumulusci.core.tasks import BaseTask
from cumulusci.core.utils import process_bool_arg
from cumulusci.core.utils import process_list_arg
from cumulusci.robotframework.utils import set_pdb_trace
from cumulusci.tasks.salesforce import BaseSalesforceTask
from cumulusci.tasks.robotframework.debugger import DebugListener


class Robot(BaseSalesforceTask):
    task_options = {
        "suites": {
            "description": 'Paths to test case files/directories to be executed similarly as when running the robot command on the command line.  Defaults to "tests" to run all tests in the tests directory',
            "required": True,
        },
        "test": {
            "description": "Run only tests matching name patterns.  Can be comma separated and use robot wildcards like *"
        },
        "include": {"description": "Includes tests with a given tag"},
        "exclude": {"description": "Excludes tests with a given tag"},
        "vars": {
            "description": "Pass values to override variables in the format VAR1:foo,VAR2:bar"
        },
        "xunit": {"description": "Set an XUnit format output file for test results"},
        "options": {
            "description": "A dictionary of options to robot.run method.  See docs here for format.  NOTE: There is no cci CLI support for this option since it requires a dictionary.  Use this option in the cumulusci.yml when defining custom tasks where you can easily create a dictionary in yaml."
        },
        "name": {"description": "Sets the name of the top level test suite"},
        "pdb": {"description": "If true, run the Python debugger when tests fail."},
        "verbose": {"description": "If true, log each keyword as it runs."},
        "debug": {
            "description": "If true, enable the `breakpoint` keyword to enable the robot debugger"
        },
    }

    def _init_options(self, kwargs):
        super(Robot, self)._init_options(kwargs)

        for option in ("test", "include", "exclude", "vars"):
            if option in self.options:
                self.options[option] = process_list_arg(self.options[option])
        if "vars" not in self.options:
            self.options["vars"] = []

        # Initialize options as a dict
        if "options" not in self.options:
            self.options["options"] = {}

        # There are potentially many robot options that are or could
        # be lists, but the only one we currently care about is the
        # listener option since we may need to append additional values
        # onto it.
        for option in ("listener",):
            if option in self.options["options"]:
                self.options["options"][option] = process_list_arg(
                    self.options["options"][option]
                )

        listeners = self.options["options"].setdefault("listener", [])
        if process_bool_arg(self.options.get("verbose")):
            listeners.append(KeywordLogger())

        if process_bool_arg(self.options.get("debug")):
            listeners.append(DebugListener())

        if process_bool_arg(self.options.get("pdb")):
            patch_statusreporter()

    def _run_task(self):
        self.options["vars"].append("org:{}".format(self.org_config.name))
        options = self.options["options"].copy()
        for option in ("test", "include", "exclude", "xunit", "name"):
            if option in self.options:
                options[option] = self.options[option]
        options["variable"] = self.options.get("vars") or []
        options["outputdir"] = os.path.relpath(
            os.path.join(self.working_path, options.get("outputdir", ".")), os.getcwd()
        )

        num_failed = robot_run(self.options["suites"], **options)
        if num_failed:
            raise RobotTestFailure("{} tests failed".format(num_failed))


class RobotTestDoc(BaseTask):
    task_options = {
        "path": {
            "description": "The path containing .robot test files",
            "required": True,
        },
        "output": {
            "description": "The output html file where the documentation will be written",
            "required": True,
        },
    }

    def _run_task(self):
        return testdoc(self.options["path"], self.options["output"])


class KeywordLogger(object):
    ROBOT_LISTENER_API_VERSION = 2

    def start_keyword(self, name, attrs):
        sys.stdout.write("  {}  {}\n".format(attrs["kwname"], "  ".join(attrs["args"])))
        sys.stdout.flush()


def patch_statusreporter():
    """Monkey patch robotframework to do postmortem debugging
    """
    from robot.running.statusreporter import StatusReporter

    orig_exit = StatusReporter.__exit__

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val and isinstance(exc_val, Exception):
            set_pdb_trace(pm=True)
        return orig_exit(self, exc_type, exc_val, exc_tb)

    StatusReporter.__exit__ = __exit__


def patch_executescript():
    # convert executeScript calls into executeAsyncScript
    # to work around an issue in chromedriver 77
    # https://bugs.chromium.org/p/chromedriver/issues/detail?id=3103
    from selenium.webdriver.remote.webdriver import WebDriver

    def execute_script(self, script, *args):
        # the last argument is the function called to say the async operation is done
        script = (
            "arguments[arguments.length - 1](function(){"
            + script
            + "}.apply(null, Array.prototype.slice.call(arguments, 0, -1)));"
        )
        return self.execute_async_script(script, *args)

    WebDriver.execute_script = execute_script


patch_executescript()



""" Tests for the Command tasks """

from unittest import mock
import unittest
from io import BytesIO

from testfixtures.popen import MockPopen
from testfixtures import Replacer

from cumulusci.core.config import BaseGlobalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.tests.utils import MockLoggerMixin

from cumulusci.tasks.command import Command
from cumulusci.tasks.command import SalesforceCommand
from cumulusci.tasks.command import CommandException


class TestCommandTask(MockLoggerMixin, unittest.TestCase):
    """ Tests for the basic command task """

    def setUp(self):
        self.global_config = BaseGlobalConfig()
        self.project_config = BaseProjectConfig(
            self.global_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig()
        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_functional_run_ls(self):
        """ Functional test that actually subprocesses and runs command.

        Checks that command either ran successfully or failed as expected
        so that it can be run on any platform. Other tests will mock
        the subprocess interaction.
        """

        self.task_config.config["options"] = {"command": "ls -la"}

        task = Command(self.project_config, self.task_config)

        # try to run the command and handle the command Exception
        # if no exception, confirm that we logged
        # if exception, confirm exception matches expectations
        try:
            task()
        except CommandException:
            self.assertTrue(any("Return code" in s for s in self.task_log["error"]))

        self.assertTrue(any("total" in s for s in self.task_log["info"]))

    def test_init__json_env(self):
        task_config = TaskConfig({"options": {"env": "{}", "command": "ls"}})
        task = Command(self.project_config, task_config)
        self.assertEqual({}, task.options["env"])

    def test_init__dict_env(self):
        task_config = TaskConfig({"options": {"env": {}, "command": "ls"}})
        task = Command(self.project_config, task_config)
        self.assertEqual({}, task.options["env"])

    def test_get_env__pass_env_false(self):
        task_config = TaskConfig({"options": {"pass_env": "False", "command": "ls"}})
        task = Command(self.project_config, task_config)
        self.assertEqual({}, task._get_env())

    def test_handle_returncode(self):
        task_config = TaskConfig({"options": {"command": "ls"}})
        task = Command(self.project_config, task_config)
        with self.assertRaises(CommandException):
            task._handle_returncode(1, BytesIO(b"err"))


class TestCommandTaskWithMockPopen(MockLoggerMixin, unittest.TestCase):
    """ Run command tasks with a mocked popen """

    def setUp(self):
        self.global_config = BaseGlobalConfig()
        self.project_config = BaseProjectConfig(
            self.global_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig()

        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

        self.Popen = MockPopen()
        self.r = Replacer()
        self.r.replace("cumulusci.tasks.command.subprocess.Popen", self.Popen)
        self.addCleanup(self.r.restore)

    def test_functional_mock_command(self):
        """ Functional test that runs a command with mocked
        popen results and checks the log.
        """

        self.task_config.config["options"] = {"command": "ls -la"}

        self.Popen.set_command("ls -la", stdout=b"testing testing 123", stderr=b"e")

        task = Command(self.project_config, self.task_config)
        task()

        self.assertTrue(any("testing testing" in s for s in self.task_log["info"]))


class TestSalesforceCommand(unittest.TestCase):
    def setUp(self):
        self.global_config = BaseGlobalConfig()
        self.project_config = BaseProjectConfig(
            self.global_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig({"options": {"command": "ls"}})
        self.org_config = OrgConfig(
            {"access_token": "TOKEN", "instance_url": "https://na01.salesforce.com"},
            "test",
        )
        self.org_config.refresh_oauth_token = mock.Mock()

    def test_update_credentials(self):
        task = SalesforceCommand(self.project_config, self.task_config, self.org_config)
        task()
        self.org_config.refresh_oauth_token.assert_called_once()

    def test_get_env(self):
        task = SalesforceCommand(self.project_config, self.task_config, self.org_config)
        env = task._get_env()
        self.assertIn("SF_ACCESS_TOKEN", env)
        self.assertIn("SF_INSTANCE_URL", env)



from unittest.mock import patch
from pathlib import Path
from ..logger import init_logger, get_gist_logger, get_rot_file_logger


class TestLogger:
    @patch("cumulusci.cli.logger.requests")
    @patch("cumulusci.cli.logger.logging")
    def test_init_logger(self, logging, requests):
        init_logger(log_requests=True)
        requests.packages.urllib3.add_stderr_logger.assert_called_once()

    @patch("cumulusci.cli.logger.Path.mkdir")
    @patch("cumulusci.cli.logger.Path.home")
    @patch("cumulusci.cli.logger.get_rot_file_logger")
    def test_get_gist_logger(self, file_logger, home, mkdir):
        home.return_value = Path("/Users/bob.ross")
        get_gist_logger()
        file_logger.assert_called_once_with(
            "stdout/stderr", Path("/Users/bob.ross/.cumulusci/logs/cci.log")
        )

    @patch("cumulusci.cli.logger.RotatingFileHandler")
    @patch("cumulusci.cli.logger.logging")
    def test_get_rot_file_logger(self, logging, rot_filehandler):

        logger_name = "The happy logger"
        path = "happy/logger/path"
        logger = get_rot_file_logger(logger_name, path)

        logging.getLogger.assert_called_once_with(logger_name)
        rot_filehandler.assert_called_once_with(path, backupCount=5, encoding="utf-8")
        logger.addHandler.assert_called_once()
        logger.setLevel.assert_called_once()



from cumulusci.tasks.github.base import BaseGithubTask


class PullRequests(BaseGithubTask):
    def _run_task(self):
        repo = self.get_repo()
        for pr in repo.pull_requests(state="open"):
            self.logger.info("#{}: {}".format(pr.number, pr.title))



from unittest import mock
import unittest
import shutil
import tempfile
import pytest
import os.path
import re
from xml.etree import ElementTree as ET

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import RobotTestFailure, TaskOptionsError
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.tasks.robotframework import Robot
from cumulusci.tasks.robotframework import RobotLibDoc
from cumulusci.tasks.robotframework import RobotTestDoc
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.robotframework.debugger import DebugListener
from cumulusci.tasks.robotframework.robotframework import KeywordLogger
from cumulusci.utils import touch

from cumulusci.tasks.robotframework.libdoc import KeywordFile


class TestRobot(unittest.TestCase):
    @mock.patch("cumulusci.tasks.robotframework.robotframework.robot_run")
    def test_run_task(self, robot_run):
        robot_run.return_value = 1
        task = create_task(Robot, {"suites": "tests", "pdb": True})
        with self.assertRaises(RobotTestFailure):
            task()

    @mock.patch("cumulusci.tasks.robotframework.robotframework.patch_statusreporter")
    def test_pdb_arg(self, patch_statusreporter):
        create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "pdb": "False",
            },
        )
        patch_statusreporter.assert_not_called()

        create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "pdb": "True",
            },
        )
        patch_statusreporter.assert_called_once()

    def test_list_args(self):
        """Verify that certain arguments are converted to lists"""
        task = create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "test": "one, two",
                "include": "foo, bar",
                "exclude": "a,  b",
                "vars": "uno, dos, tres",
            },
        )
        for option in ("test", "include", "exclude", "vars"):
            assert isinstance(task.options[option], list)

    def test_default_listeners(self):
        # first, verify that not specifying any listener options
        # results in no listeners...
        task = create_task(
            Robot, {"suites": "test"}  # required, or the task will raise an exception
        )
        assert len(task.options["options"]["listener"]) == 0

        # next, make sure that if we specify the options with a Falsy
        # string, the option is properly treated like a boolean
        task = create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "verbose": "False",
                "debug": "False",
            },
        )
        assert len(task.options["options"]["listener"]) == 0

    def test_debug_option(self):
        """Verify that setting debug to True attaches the appropriate listener"""
        task = create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "debug": "True",
            },
        )
        listener_classes = [
            listener.__class__ for listener in task.options["options"]["listener"]
        ]
        self.assertIn(
            DebugListener, listener_classes, "DebugListener was not in task options"
        )

    def test_verbose_option(self):
        """Verify that setting verbose to True attaches the appropriate listener"""
        task = create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "verbose": "True",
            },
        )
        listener_classes = [
            listener.__class__ for listener in task.options["options"]["listener"]
        ]
        self.assertIn(
            KeywordLogger, listener_classes, "KeywordLogger was not in task options"
        )

    def test_user_defined_listeners_option(self):
        """Verify that our listeners don't replace user-defined listeners"""
        task = create_task(
            Robot,
            {
                "suites": "test",  # required, or the task will raise an exception
                "debug": "True",
                "verbose": "True",
                "options": {"listener": ["FakeListener.py"]},
            },
        )
        listener_classes = [
            listener.__class__ for listener in task.options["options"]["listener"]
        ]
        self.assertIn("FakeListener.py", task.options["options"]["listener"])
        self.assertIn(DebugListener, listener_classes)
        self.assertIn(KeywordLogger, listener_classes)


class TestRobotTestDoc(unittest.TestCase):
    @mock.patch("cumulusci.tasks.robotframework.robotframework.testdoc")
    def test_run_task(self, testdoc):
        task = create_task(RobotTestDoc, {"path": ".", "output": "out"})
        task()
        testdoc.assert_called_once_with(".", "out")


class TestRobotLibDoc(MockLoggerMixin, unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(dir=".")
        self.task_config = TaskConfig()
        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages
        self.datadir = os.path.dirname(__file__)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_output_directory_not_exist(self):
        """Verify we catch an error if the output directory doesn't exist"""
        path = os.path.join(self.datadir, "TestLibrary.py")
        output = os.path.join(self.tmpdir, "bogus", "index.html")
        task = create_task(RobotLibDoc, {"path": path, "output": output})
        # on windows, the output path may have backslashes which needs
        # to be protected in the expected regex
        expected = r"Unable to create output file '{}' (.*)".format(re.escape(output))
        with pytest.raises(TaskOptionsError, match=expected):
            task()

    def test_validate_filenames(self):
        """Verify that we catch bad filenames early"""
        expected = "Unable to find the following input files: 'bogus.py', 'bogus.robot'"
        output = os.path.join(self.tmpdir, "index.html")
        with pytest.raises(TaskOptionsError, match=expected):
            create_task(RobotLibDoc, {"path": "bogus.py,bogus.robot", "output": output})

        # there's a special path through the code if only one filename is bad...
        expected = "Unable to find the input file 'bogus.py'"
        with pytest.raises(TaskOptionsError, match=expected):
            create_task(RobotLibDoc, {"path": "bogus.py", "output": output})

    def test_task_log(self):
        """Verify that the task prints out the name of the output file"""
        path = os.path.join(self.datadir, "TestLibrary.py")
        output = os.path.join(self.tmpdir, "index.html")
        task = create_task(RobotLibDoc, {"path": path, "output": output})
        task()
        assert "created {}".format(output) in self.task_log["info"]
        assert os.path.exists(output)

    def test_comma_separated_list_of_files(self):
        """Verify that we properly parse a comma-separated list of files"""
        path = "{},{}".format(
            os.path.join(self.datadir, "TestLibrary.py"),
            os.path.join(self.datadir, "TestResource.robot"),
        )
        output = os.path.join(self.tmpdir, "index.html")
        task = create_task(RobotLibDoc, {"path": path, "output": output})
        task()
        assert os.path.exists(output)
        assert len(task.result["files"]) == 2

    def test_glob_patterns(self):
        output = os.path.join(self.tmpdir, "index.html")
        path = os.path.join(self.datadir, "*Library.py")
        task = create_task(RobotLibDoc, {"path": path, "output": output})
        task()
        assert os.path.exists(output)
        assert len(task.result["files"]) == 1
        assert task.result["files"][0] == os.path.join(self.datadir, "TestLibrary.py")

    def test_remove_duplicates(self):
        output = os.path.join(self.tmpdir, "index.html")
        path = os.path.join(self.datadir, "*Library.py")
        task = create_task(RobotLibDoc, {"path": [path, path], "output": output})
        task()
        assert len(task.result["files"]) == 1
        assert task.result["files"][0] == os.path.join(self.datadir, "TestLibrary.py")

    def test_creates_output(self):
        path = os.path.join(self.datadir, "TestLibrary.py")
        output = os.path.join(self.tmpdir, "index.html")
        task = create_task(RobotLibDoc, {"path": path, "output": output})
        task()
        assert "created {}".format(output) in self.task_log["info"]
        assert os.path.exists(output)

    def test_pageobject(self):
        """Verify that we can parse a page object file"""
        path = os.path.join(self.datadir, "TestPageObjects.py")
        output = os.path.join(self.tmpdir, "index.html")
        task = create_task(RobotLibDoc, {"path": path, "output": output})
        task()
        assert "created {}".format(output) in self.task_log["info"]
        assert os.path.exists(output)


class TestRobotLibDocKeywordFile(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(dir=".")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_existing_file(self):
        path = os.path.join(self.tmpdir, "keywords.py")
        touch(path)
        kwfile = KeywordFile(path)
        assert kwfile.filename == "keywords.py"
        assert kwfile.path == path
        assert kwfile.keywords == {}

    def test_file_as_module(self):
        kwfile = KeywordFile("cumulusci.robotframework.Salesforce")
        assert kwfile.filename == "Salesforce"
        assert kwfile.path == "cumulusci.robotframework.Salesforce"
        assert kwfile.keywords == {}

    def test_add_keyword(self):
        kwfile = KeywordFile("test.TestLibrary")
        kwfile.add_keywords("the documentation...", ("Detail", "Contact"))
        assert len(kwfile.keywords) == 1
        assert kwfile.keywords[("Detail", "Contact")] == "the documentation..."


class TestRobotLibDocOutput(unittest.TestCase):
    """Tests for the generated robot keyword documentation"""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(dir=".")
        self.datadir = os.path.dirname(__file__)
        path = [
            os.path.join(self.datadir, "TestLibrary.py"),
            os.path.join(self.datadir, "TestResource.robot"),
        ]
        output = os.path.join(self.tmpdir, "index.html")
        self.task = create_task(
            RobotLibDoc,
            {"path": path, "output": output, "title": "Keyword Documentation, yo."},
        )
        self.task()
        docroot = ET.parse(output).getroot()
        self.html_body = docroot.find("body")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_output_title(self):
        """Verify the document has the expected title"""
        title_element = self.html_body.find(
            ".//div[@class='header']/h1[@class='title']"
        )
        assert title_element is not None
        assert title_element.text.strip() == "Keyword Documentation, yo."

    def test_formatted_documentation(self):
        """Verify that markup in the documentation is rendered as html"""
        doc_element = self.html_body.find(
            ".//tr[@id='TestLibrary.py.Library Keyword One']//td[@class='kwdoc']"
        )
        doc_html = str(ET.tostring(doc_element, method="html").strip())
        # we could just do an assert on the full markup of the
        # element, but it seems likely that could fail for benign
        # regions (extra whitespace, for example). So we'll just make
        # sure there are a couple of expected formatted elements.
        assert "<b>bold</b>" in doc_html
        assert "<i>italics</i>" in doc_html

    def test_output_sections(self):
        """Verify that the output has a section for each file"""
        sections = self.html_body.findall(".//div[@class='file']")
        section_titles = [
            x.find(".//div[@class='file-header']/h2").text for x in sections
        ]
        assert len(sections) == 2, "expected to find 2 sections, found {}".format(
            len(sections)
        )
        assert section_titles == ["TestLibrary.py", "TestResource.robot"]

    def test_output_keywords(self):
        """Verify that all keywords in the libraries are represented in the output file"""
        keyword_rows = self.html_body.findall(".//tr[@class='kwrow']")
        keywords = [x.find(".//td[@class='kwname']") for x in keyword_rows]
        keyword_names = [x.text.strip() for x in keywords]
        assert len(keywords) == 4
        assert keyword_names == [
            "Library Keyword One",
            "Library Keyword Two",
            "Resource keyword one",
            "Resource keyword two",
        ]


class TestLibdocPageObjects(unittest.TestCase):
    """Tests for generating docs for page objects"""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(dir=".")
        self.datadir = os.path.dirname(__file__)
        path = [os.path.join(self.datadir, "TestPageObjects.py")]
        self.output = os.path.join(self.tmpdir, "index.html")
        self.task = create_task(
            RobotLibDoc,
            {"path": path, "output": self.output, "title": "Keyword Documentation, yo"},
        )
        self.task()

        self.docroot = ET.parse(self.output).getroot()
        self.html_body = self.docroot.find("body")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_file_title(self):
        """Verify that the TITLE in the file is added to the generated html"""
        # The page object file has a TITLE attribute and a docstring;
        # make sure they are picked up.
        title_element = self.html_body.find(".//div[@class='file-header']/h2")
        assert title_element.text == "This is the title"

    def test_file_description(self):
        """Verify that the docstring in the file is added to the generated html"""
        file_doc_element = self.html_body.find(
            ".//div[@class='pageobject-file-description']"
        )
        description = ET.tostring(file_doc_element).decode("utf-8").strip()
        assert (
            description
            == '<div class="pageobject-file-description"><p>this is the docstring</p></div>'
        )

    def test_pageobject_sections(self):
        # the TestPageObjects.py file has two page objects,
        # one with two keywords and one with three
        sections = self.html_body.findall(".//div[@class='pageobject-header']")
        assert len(sections) == 2

    def test_pageobject_docstring(self):
        section = self.html_body.find(".//div[@pageobject='Detail-Something__c']")
        description = section.find("div[@class='description']")
        expected = '<div class="description" title="Description"><p>Description of SomethingDetailPage</p></div>'
        actual = ET.tostring(description).decode("utf-8").strip()
        assert actual == expected

        section = self.html_body.find(".//div[@pageobject='Listing-Something__c']")
        description = section.find("div[@class='description']")
        expected = '<div class="description" title="Description"><p>Description of SomethingListingPage</p></div>'
        actual = ET.tostring(description).decode("utf-8").strip()
        assert actual == expected



from cumulusci.robotframework import locators_47

lex_locators = locators_47.lex_locators.copy()

lex_locators["record"]["header"]["field_value"] = (
    "//records-lwc-highlights-panel"
    "//force-highlights-details-item"
    "[.//*[contains(@class, 'slds-text-title') and text()='{}']]"
    "//p[contains(@class, 'fieldComponent')]//*[text()]"
)

lex_locators["record"]["header"]["field_value_link"] = (
    "//records-lwc-highlights-panel"
    "//force-highlights-details-item"
    "[.//*[contains(@class, 'slds-text-title') and text()='{}']]"
    "//p[contains(@class, 'fieldComponent')]//a[text()]"
)

lex_locators["record"]["related"]["link"] = (
    "//article[contains(@class, 'forceRelatedListCardDesktop')]"
    "[.//img]"
    "[.//span[@title='{}']]"
    "//*[text()='{}']"
)

lex_locators["app_launcher"]["menu"] = "//div[contains(@class, 'appLauncherMenu')]"
lex_locators["app_launcher"]["view_all"] = (
    "//div[contains(@class, 'appLauncherMenu')]" "//button[text()='View All']"
)
lex_locators["app_launcher"][
    "app_link"
] = "//one-app-launcher-modal//one-app-launcher-app-tile//a[.='{}']"
lex_locators["app_launcher"][
    "tab_link"
] = "//one-app-launcher-modal//one-app-launcher-tab-item//a[.='{}']"

lex_locators["object"]["field_lookup_link"] = "//*[@role='option'][.//*[@title='{}']]"

# NPSP tests were failing because the New Opportunity modal no longer
# is using the class modal-footer.
lex_locators["modal"]["button"] = (
    "//div[contains(@class,'uiModal')]"
    "//div[contains(@class, 'modal-footer') or contains(@class, 'inlineFooter')]"
    # I don't like using an '*' here, but the link might be an <a> or it
    # might be a <span>.
    "//button[.//*[text()='{}']]"
)



import os
import sys

__import__("pkg_resources").declare_namespace("cumulusci")

__version__ = "3.7.0"

__location__ = os.path.dirname(os.path.realpath(__file__))

if sys.version_info < (3, 6):  # pragma: no cover
    raise Exception("CumulusCI requires Python 3.6+.")



from unittest import mock
import responses
import unittest
from cumulusci.tasks.salesforce import SetTDTMHandlerStatus
from cumulusci.core.exceptions import TaskOptionsError, CumulusCIException
from .util import create_task


class test_trigger_handlers(unittest.TestCase):
    def test_init_options(self):
        task = create_task(
            SetTDTMHandlerStatus, {"handlers": ["TestTDTM"], "active": False}
        )

        assert task.options["handlers"] == ["TestTDTM"]

        with self.assertRaises(TaskOptionsError):
            task = create_task(
                SetTDTMHandlerStatus, {"handlers": ["TestTDTM"], "restore": True}
            )

    @responses.activate
    def test_missing_handler_object(self):
        task = create_task(
            SetTDTMHandlerStatus, {"handlers": ["TestTDTM"], "active": False}
        )
        task.api_version = "47.0"
        responses.add(
            method="GET",
            url=task.org_config.instance_url + "/services/data/v47.0/sobjects",
            json={"sobjects": [{"name": "t__c"}]},
            status=200,
        )

        with self.assertRaises(CumulusCIException):
            task()

    @responses.activate
    def test_set_status(self):
        task = create_task(
            SetTDTMHandlerStatus,
            {"handlers": ["TestTDTM"], "active": False, "namespace": "npsp"},
        )
        task.api_version = "47.0"
        responses.add(
            method="GET",
            url=task.org_config.instance_url + "/services/data/v47.0/sobjects",
            json={"sobjects": [{"name": "Trigger_Handler__c"}]},
            status=200,
        )
        responses.add(
            method="GET",
            url=task.org_config.instance_url
            + "/services/data/v47.0/query/?q=SELECT+Id%2C+Class__c%2C+Object__c%2C+Active__c+FROM+Trigger_Handler__c",
            status=200,
            json={
                "records": [
                    {
                        "Id": "000000000000000",
                        "Class__c": "TestTDTM",
                        "Object__c": "Test__c",
                        "Active__c": True,
                    },
                    {
                        "Id": "000000000000001",
                        "Class__c": "Test",
                        "Object__c": "Test__c",
                        "Active__c": True,
                    },
                ]
            },
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/Trigger_Handler__c/000000000000000",
            status=204,
            json={"Active__c": False},
        )
        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_set_status__all_handlers(self):
        task = create_task(SetTDTMHandlerStatus, {"active": False, "namespace": "npsp"})
        task.api_version = "47.0"
        responses.add(
            method="GET",
            url=task.org_config.instance_url + "/services/data/v47.0/sobjects",
            json={"sobjects": [{"name": "Trigger_Handler__c"}]},
            status=200,
        )
        responses.add(
            method="GET",
            url=task.org_config.instance_url
            + "/services/data/v47.0/query/?q=SELECT+Id%2C+Class__c%2C+Object__c%2C+Active__c+FROM+Trigger_Handler__c",
            status=200,
            json={
                "records": [
                    {
                        "Id": "000000000000000",
                        "Class__c": "TestTDTM",
                        "Object__c": "Test__c",
                        "Active__c": True,
                    },
                    {
                        "Id": "000000000000001",
                        "Class__c": "Test",
                        "Object__c": "Test__c",
                        "Active__c": True,
                    },
                ]
            },
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/Trigger_Handler__c/000000000000000",
            status=204,
            json={"Active__c": False},
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/Trigger_Handler__c/000000000000001",
            status=204,
            json={"Active__c": False},
        )
        task()

        assert len(responses.calls) == 4

    @responses.activate
    def test_set_status_namespaced(self):
        task = create_task(
            SetTDTMHandlerStatus,
            {"handlers": ["Test__c:TestTDTM"], "active": False, "namespace": "npsp"},
        )
        task.api_version = "47.0"
        responses.add(
            method="GET",
            url=task.org_config.instance_url + "/services/data/v47.0/sobjects",
            json={"sobjects": [{"name": "npsp__Trigger_Handler__c"}]},
            status=200,
        )
        responses.add(
            method="GET",
            url=task.org_config.instance_url
            + "/services/data/v47.0/query/?q=SELECT+Id%2C+npsp__Class__c%2C+npsp__Object__c%2C+npsp__Active__c+FROM+npsp__Trigger_Handler__c",
            status=200,
            json={
                "records": [
                    {
                        "Id": "000000000000000",
                        "npsp__Class__c": "TestTDTM",
                        "npsp__Object__c": "Test__c",
                        "npsp__Active__c": True,
                    },
                    {
                        "Id": "000000000000001",
                        "npsp__Class__c": "Test",
                        "npsp__Object__c": "Test__c",
                        "npsp__Active__c": True,
                    },
                ]
            },
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/npsp__Trigger_Handler__c/000000000000000",
            status=204,
            json={"npsp__Active__c": False},
        )
        task()

        assert len(responses.calls) == 3

    @responses.activate
    def test_restore(self):
        task = create_task(
            SetTDTMHandlerStatus,
            {"restore": True, "restore_file": "resto.yml", "namespace": "npsp"},
        )
        task.api_version = "47.0"
        responses.add(
            method="GET",
            url=task.org_config.instance_url + "/services/data/v47.0/sobjects",
            json={"sobjects": [{"name": "npsp__Trigger_Handler__c"}]},
            status=200,
        )
        responses.add(
            method="GET",
            url=task.org_config.instance_url
            + "/services/data/v47.0/query/?q=SELECT+Id%2C+npsp__Class__c%2C+npsp__Object__c%2C+npsp__Active__c+FROM+npsp__Trigger_Handler__c",
            status=200,
            json={
                "records": [
                    {
                        "Id": "000000000000000",
                        "npsp__Class__c": "TestTDTM",
                        "npsp__Object__c": "Test__c",
                        "npsp__Active__c": False,
                    },
                    {
                        "Id": "000000000000001",
                        "npsp__Class__c": "Test",
                        "npsp__Object__c": "Test__c",
                        "npsp__Active__c": False,
                    },
                ]
            },
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/npsp__Trigger_Handler__c/000000000000000",
            status=204,
            json={"npsp__Active__c": True},
        )

        m = mock.mock_open(read_data="'Test__c:TestTDTM': True")
        with mock.patch("builtins.open", m):
            with mock.patch("os.remove") as rm_patch:
                task()

                m.assert_any_call("resto.yml", "r")
                rm_patch.assert_called_once_with("resto.yml")

        assert len(responses.calls) == 3

    @responses.activate
    def test_create_restore_file(self):
        task = create_task(
            SetTDTMHandlerStatus,
            {
                "handlers": ["Test__c"],
                "active": False,
                "restore_file": "resto.yml",
                "namespace": "npsp",
            },
        )
        task.api_version = "47.0"
        responses.add(
            method="GET",
            url=task.org_config.instance_url + "/services/data/v47.0/sobjects",
            json={"sobjects": [{"name": "Trigger_Handler__c"}]},
            status=200,
        )
        responses.add(
            method="GET",
            url=task.org_config.instance_url
            + "/services/data/v47.0/query/?q=SELECT+Id%2C+Class__c%2C+Object__c%2C+Active__c+FROM+Trigger_Handler__c",
            status=200,
            json={
                "records": [
                    {
                        "Id": "000000000000000",
                        "Class__c": "TestTDTM",
                        "Object__c": "Test__c",
                        "Active__c": True,
                    },
                    {
                        "Id": "000000000000001",
                        "Class__c": "Test",
                        "Object__c": "Test__c",
                        "Active__c": True,
                    },
                ]
            },
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/Trigger_Handler__c/000000000000000",
            status=204,
            json={"Active__c": False},
        )
        responses.add(
            method="PATCH",
            url=task.org_config.instance_url
            + "/services/data/v47.0/sobjects/Trigger_Handler__c/000000000000001",
            status=204,
            json={"Active__c": False},
        )
        m = mock.mock_open()
        with mock.patch("builtins.open", m):
            with mock.patch("yaml.safe_dump") as yaml_mock:
                task()

            m.assert_any_call("resto.yml", "w")
            yaml_mock.assert_called_once_with(
                {"Test__c:TestTDTM": True, "Test__c:Test": True}, m.return_value
            )

        assert len(responses.calls) == 4



from robot.libraries.BuiltIn import BuiltIn


class BasePage:
    _object_name = None

    def __init__(self, object_name=None):
        if object_name:
            self._object_name = object_name

    def _wait_to_appear(self, timeout=None):
        """This function is called by the keyword 'wait for page object to appear'.
        Custom page objects can override this to verify that the object is visible.
        """
        raise Exception("Unable to wait for this page object")

    def _remove_from_library_search_order(self):
        """Remove the current page object from robot's library search order

        Note: robot doesn't provide a way to completely unload a
        library during execution. However, this at least makes sure that
        all other libraries will have priority over this one.
        """
        order = list(self.builtin.set_library_search_order())
        if self._libname in order:
            order.remove(self._libname)
            self.builtin.set_library_search_order(*order)
            self.builtin.log("new search order: {}".format(order), "DEBUG")

    @property
    def object_name(self):
        object_name = self._object_name

        # the length check is to skip objects from a different namespace
        # like foobar__otherpackageobject__c
        if object_name is not None:
            parts = object_name.split("__")
            if len(parts) == 2 and parts[-1] == "c":
                # get_namespace_prefix already takes care of returning an actual
                # prefix or an empty string depending on whether the package is managed
                object_name = "{}{}".format(
                    self.cumulusci.get_namespace_prefix(), object_name
                )
        return object_name

    @property
    def builtin(self):
        """Returns an instance of robot framework's BuiltIn library"""
        return BuiltIn()

    @property
    def cumulusci(self):
        """Returns the instance of the imported CumulusCI library"""
        return self.builtin.get_library_instance("cumulusci.robotframework.CumulusCI")

    @property
    def salesforce(self):
        """Returns the instance of the imported Salesforce library"""
        return self.builtin.get_library_instance("cumulusci.robotframework.Salesforce")

    @property
    def selenium(self):
        """Returns the instance of the imported SeleniumLibrary library"""
        return self.builtin.get_library_instance("SeleniumLibrary")

    def _get_object(self, **kwargs):
        """Get the object associated with the given keyword arguments.

        This performs a salesforce query. It will raise an exception unless
        exactly one result is returned from the query.
        """
        results = self.salesforce.salesforce_query(self.object_name, **kwargs)
        if len(results) == 0:
            human_friendly_args = ", ".join(
                ["{}={}".format(key, kwargs[key]) for key in kwargs]
            )
            raise Exception(
                "no {} matches {}".format(self.object_name, human_friendly_args)
            )
        elif len(results) > 1:
            raise Exception("Query returned {} objects".format(len(results)))
        else:
            return results[0]

    def log_current_page_object(self):
        """Logs the name of the current page object

        The current page object is also returned as an object
        """
        polib = self.builtin.get_library_instance(
            "cumulusci.robotframework.PageObjects"
        )
        if polib.current_page_object is None:
            # this should not be possible, since this keyword is only available
            # if you've loaded a page object. Still, better safe then sorry.
            self.builtin.log("no page object has been loaded")
            return None
        else:
            pobj = polib.current_page_object

            self.builtin.log(
                "current page object: {}".format(polib.current_page_object)
            )
            return pobj



from functools import lru_cache
from faker import Faker

from jinja2 import Template


class StringGenerator:
    """ Sometimes in templates you want a reference to a variable to
        call a function.

        For example:

        >>> x = template_utils.StringGenerator(datetime.today().isoformat)
        >>> print(f"{x}")
        2019-09-23T11:49:01.994453

        >>> x = template_utils.StringGenerator(lambda:str(random.random()))
        >>> print(f"{x}")
        0.795273959965055
        >>> print(f"{x}")
        0.053061903749985206
         """

    def __init__(self, func):
        self.func = func

    def __str__(self):
        return self.func()

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class FakerTemplateLibrary:
    """A Jinja template library to add the faker.xyz objects to templates"""

    _faker = None

    def __init__(self, locale=None):
        self.locale = locale
        self.faker = Faker(self.locale)

    def __getattr__(self, name):
        return StringGenerator(
            lambda *args, **kwargs: self.faker.format(name, *args, **kwargs)
        )


faker_template_library = FakerTemplateLibrary()

Template = lru_cache(512)(Template)


def format_str(value, variables=None, fake=faker_template_library):
    variables = variables or {}
    if isinstance(value, str) and "{" in value:
        value = Template(value).render(fake=fake, **variables)

    return value



""" CLI logger """
import logging
from logging.handlers import RotatingFileHandler
import os
import sys
import requests
from pathlib import Path

import coloredlogs

try:
    import colorama
except ImportError:
    # coloredlogs only installs colorama on Windows
    pass


def init_logger(log_requests=False):
    """ Initialize the logger """

    logger = logging.getLogger(__name__.split(".")[0])
    for handler in logger.handlers:  # pragma: no cover
        logger.removeHandler(handler)

    if os.name == "nt" and "colorama" in sys.modules:  # pragma: no cover
        colorama.init()

    formatter = coloredlogs.ColoredFormatter(fmt="%(asctime)s: %(message)s")
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if log_requests:
        requests.packages.urllib3.add_stderr_logger()


def get_gist_logger():
    """Determines the appropriate filepath for logfile
    and name for the logger. Returns a logger with
    RotatingFileHandler attached."""
    logfile_dir = Path.home() / ".cumulusci" / "logs"
    logfile_dir.mkdir(parents=True, exist_ok=True)
    logfile_path = logfile_dir / "cci.log"

    return get_rot_file_logger("stdout/stderr", logfile_path)


def get_rot_file_logger(name, path):
    """Returns a logger with a rotating file handler"""
    logger = logging.getLogger(name)

    handler = RotatingFileHandler(path, backupCount=5, encoding="utf-8")
    handler.doRollover()  # rollover existing log files
    handler.terminator = ""  # click.echo already adds a newline
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return logger



import simple_salesforce
from cumulusci import __version__
from cumulusci.core.exceptions import ServiceNotConfigured, ServiceNotValid
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

CALL_OPTS_HEADER_KEY = "Sforce-Call-Options"


def get_simple_salesforce_connection(project_config, org_config, api_version=None):
    # Retry on long-running metadeploy jobs
    retries = Retry(total=5, status_forcelist=(502, 503, 504), backoff_factor=0.3)
    adapter = HTTPAdapter(max_retries=retries)

    sf = simple_salesforce.Salesforce(
        instance_url=org_config.instance_url,
        session_id=org_config.access_token,
        version=api_version or project_config.project__package__api_version,
    )
    try:
        app = project_config.keychain.get_service("connectedapp")
        client_name = app.client_id
    except (ServiceNotValid, ServiceNotConfigured):
        client_name = "CumulusCI/{}".format(__version__)

    sf.headers.setdefault(CALL_OPTS_HEADER_KEY, "client={}".format(client_name))
    sf.session.mount("http://", adapter)
    sf.session.mount("https://", adapter)

    return sf



from cumulusci.tasks.metadata_etl.base import (
    BaseMetadataETLTask,
    BaseMetadataSynthesisTask,
    BaseMetadataTransformTask,
    MetadataSingleEntityTransformTask,
    get_new_tag_index,
    MD,
)
from cumulusci.tasks.metadata_etl.layouts import AddRelatedLists
from cumulusci.tasks.metadata_etl.permissions import AddPermissionSetPermissions
from cumulusci.tasks.metadata_etl.value_sets import AddValueSetEntries
from cumulusci.tasks.metadata_etl.sharing import SetOrgWideDefaults

flake8 = (
    BaseMetadataETLTask,
    BaseMetadataSynthesisTask,
    BaseMetadataTransformTask,
    MetadataSingleEntityTransformTask,
    AddRelatedLists,
    AddPermissionSetPermissions,
    AddValueSetEntries,
    SetOrgWideDefaults,
    get_new_tag_index,
    MD,
)



from lxml import etree

from cumulusci.core.utils import process_list_arg
from cumulusci.tasks.metadata_etl import (
    MetadataSingleEntityTransformTask,
    get_new_tag_index,
    MD,
)


class AddRelatedLists(MetadataSingleEntityTransformTask):
    entity = "Layout"
    task_options = {
        "related_list": {
            "description": "Name of the Related List to include",
            "required": True,
        },
        "fields": {
            "description": "Array of field API names to include in the related list",
            "required": False,
        },
        "exclude_buttons": {
            "description": "Array of button names to suppress from the related list"
        },
        "custom_buttons": {
            "description": "Array of button names to add to the related list"
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _transform_entity(self, metadata, api_name):
        new_related_list_index = get_new_tag_index(metadata, "relatedLists")
        related_list = self._inject_namespace(self.options["related_list"])
        existing_related_lists = metadata.findall(
            f".//{MD}relatedLists[{MD}relatedList='{related_list}']"
        )

        if existing_related_lists:
            return None

        self._create_new_related_list(
            metadata, api_name, related_list, new_related_list_index
        )

        return metadata

    def _create_new_related_list(self, metadata, api_name, related_list, index):
        self.logger.info(f"Adding Related List {related_list} to {api_name}")

        fields = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("fields", []))
        ]
        exclude_buttons = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("exclude_buttons", []))
        ]
        custom_buttons = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("custom_buttons", []))
        ]

        elem = etree.Element(f"{MD}relatedLists")
        metadata.getroot().insert(index, elem)

        for f in fields:
            elem_field = etree.SubElement(elem, f"{MD}fields")
            elem_field.text = f

        for button in exclude_buttons:
            elem_button = etree.SubElement(elem, f"{MD}excludeButtons")
            elem_button.text = button

        for button in custom_buttons:
            elem_button = etree.SubElement(elem, f"{MD}customButtons")
            elem_button.text = button

        elem_related_list = etree.SubElement(elem, f"{MD}relatedList")
        elem_related_list.text = related_list



"""
python interface to the Salesforce Metadata API
based on mrbelvedere/mpinstaller/mdapi.py
"""

# TO DO
#   - add docstrings
#   - look at https://github.com/rholder/retrying

import base64
import http.client
import re
import time
from collections import defaultdict
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
from zipfile import ZipFile
import io

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from cumulusci.salesforce_api import soap_envelopes
from cumulusci.core.exceptions import ApexTestException
from cumulusci.utils import zip_subfolder, parse_api_datetime
from cumulusci.salesforce_api.exceptions import MetadataComponentFailure
from cumulusci.salesforce_api.exceptions import MetadataParseError
from cumulusci.salesforce_api.exceptions import MetadataApiError

from urllib3.contrib import pyopenssl

pyopenssl.extract_from_urllib3()

retry_policy = Retry(backoff_factor=0.3)


class BaseMetadataApiCall(object):
    check_interval = 1
    soap_envelope_start = None
    soap_envelope_status = None
    soap_envelope_result = None
    soap_action_start = None
    soap_action_status = None
    soap_action_result = None

    def __init__(self, task, api_version=None):
        # the cumulusci context object contains logger, oauth, ID, secret, etc
        self.task = task
        self.status = None
        self.check_num = 1
        self.api_version = (
            api_version
            if api_version
            else task.project_config.project__package__api_version
        )

    def __call__(self):
        self.task.logger.info("Pending")
        response = self._get_response()
        if self.status != "Failed":
            try:
                return self._process_response(response)
            except Exception as e:
                raise MetadataParseError(
                    f"Could not process MDAPI response: {str(e)}", response=response
                )

    def _build_endpoint_url(self):
        # Parse org id from id which ends in /ORGID/USERID
        org_id = self.task.org_config.org_id
        # If "My Domain" is configured in the org, the instance_url needs to be
        # parsed differently
        instance_url = self.task.org_config.instance_url
        if instance_url.find(".my.salesforce.com") != -1:
            # Parse instance_url with My Domain configured
            # URL will be in the format
            # https://name--name.na11.my.salesforce.com and should be
            # https://na11.salesforce.com
            instance_url = re.sub(
                r"https://.*\.(\w+)\.my\.salesforce\.com",
                r"https://\1.salesforce.com",
                instance_url,
            )
        # Build the endpoint url from the instance_url
        endpoint = f"{instance_url}/services/Soap/m/{self.api_version}/{org_id}"
        return endpoint

    def _build_envelope_result(self):
        if self.soap_envelope_result:
            return self.soap_envelope_result.format(process_id=self.process_id)

    def _build_envelope_start(self):
        if self.soap_envelope_start:
            return self.soap_envelope_start.format(api_version=self.api_version)

    def _build_envelope_status(self):
        if self.soap_envelope_status:
            return self.soap_envelope_status.format(process_id=self.process_id)

    def _build_headers(self, action, message):
        return {
            "Content-Type": "text/xml; charset=UTF-8",
            "Content-Length": str(len(message)),
            "SOAPAction": action,
        }

    def _call_mdapi(self, headers, envelope, refresh=None):
        # Insert the session id
        session_id = self.task.org_config.access_token
        auth_envelope = envelope.replace("###SESSION_ID###", session_id)
        session = requests.Session()
        http_adapter = HTTPAdapter(max_retries=retry_policy)
        session.mount("https://", http_adapter)
        response = session.post(
            self._build_endpoint_url(),
            headers=headers,
            data=auth_envelope.encode("utf-8"),
        )
        faultcode = parseString(response.content).getElementsByTagName("faultcode")
        # refresh = False can be passed to prevent a loop if refresh fails
        if refresh is None:
            refresh = True
        if faultcode:
            return self._handle_soap_error(headers, envelope, refresh, response)
        return response

    def _get_element_value(self, dom, tag):
        result = dom.getElementsByTagName(tag)
        if result and result[0].firstChild:
            return result[0].firstChild.nodeValue

    def _get_check_interval(self):
        return self.check_interval * ((self.check_num // 3) + 1)

    def _get_response(self):
        if not self.soap_envelope_start:
            raise NotImplementedError("No soap_start template was provided")
        # Start the call
        envelope = self._build_envelope_start()
        headers = self._build_headers(self.soap_action_start, envelope)
        response = self._call_mdapi(headers, envelope)
        # If no status envelope is configured, return the response directly
        if not self.soap_envelope_status:
            return response
        # Process the response to set self.process_id with the process id
        # started
        response = self._process_response_start(response)
        # Check the status if configured
        if self.soap_envelope_status:
            while self.status not in ["Done", "Failed"]:
                # Check status in a loop until done
                envelope = self._build_envelope_status()
                headers = self._build_headers(self.soap_action_status, envelope)
                response = self._call_mdapi(headers, envelope)
                response = self._process_response_status(response)

                # start increasing the check interval progressively to handle long pending jobs
                check_interval = self._get_check_interval()
                self.check_num += 1

                time.sleep(check_interval)
            # Fetch the final result and return
            if self.soap_envelope_result:
                envelope = self._build_envelope_result()
                headers = self._build_headers(self.soap_action_result, envelope)
                response = self._call_mdapi(headers, envelope)
            else:
                return response
        return response

    def _handle_soap_error(self, headers, envelope, refresh, response):
        resp_xml = parseString(response.content)
        faultcode = resp_xml.getElementsByTagName("faultcode")
        if faultcode:
            faultcode = faultcode[0].firstChild.nodeValue
        faultstring = resp_xml.getElementsByTagName("faultstring")
        if faultstring:
            faultstring = faultstring[0].firstChild.nodeValue
        else:
            faultstring = response.text
        if (
            faultcode == "sf:INVALID_SESSION_ID"
            and self.task.org_config
            and self.task.org_config.refresh_token
        ):
            # Attempt to refresh token and recall request
            if refresh:
                self.task.org_config.refresh_oauth_token(
                    self.task.project_config.keychain
                )
                return self._call_mdapi(headers, envelope, refresh=False)
        # Log the error
        message = f"{faultcode}: {faultstring}"
        self._set_status("Failed", message)
        raise MetadataApiError(message, response)

    def _process_response(self, response):
        return response.text

    def _process_response_start(self, response):
        if response.status_code == http.client.INTERNAL_SERVER_ERROR:
            raise MetadataApiError(
                f"HTTP ERROR {response.status_code}: {response.text}", response
            )
        ids = parseString(response.content).getElementsByTagName("id")
        if ids:
            self.process_id = ids[0].firstChild.nodeValue
        return response

    def _process_response_status(self, response):
        if response.status_code == http.client.INTERNAL_SERVER_ERROR:
            raise MetadataApiError(
                "HTTP ERROR {}: {}".format(response.status_code, response.text),
                response,
            )
        resp_xml = parseString(response.content)
        done = resp_xml.getElementsByTagName("done")
        if done:
            if done[0].firstChild.nodeValue == "true":
                self._set_status("Done")
            else:
                state_detail = resp_xml.getElementsByTagName("stateDetail")
                if state_detail:
                    log = state_detail[0].firstChild.nodeValue
                    self._set_status("InProgress", log)
                    self.check_num = 1
                elif self.status == "InProgress":
                    self.check_num = 1
                    self._set_status(
                        "InProgress",
                        f"next check in {self._get_check_interval()} seconds",
                    )
                else:
                    self._set_status(
                        "Pending", f"next check in {self._get_check_interval()} seconds"
                    )
        else:
            # If no done element was in the xml, fail logging the entire SOAP
            # envelope as the log
            self._set_status("Failed", response.text, response=response)
        return response

    def _set_status(self, status, log=None, level=None, response=None):
        if not level:
            level = "info"
            if status == "Failed":
                level = "error"
        logger = getattr(self.task.logger, level)
        self.status = status
        if log:
            logger(f"[{status}]: {log}")
        else:
            logger(f"[{status}]")


class ApiRetrieveUnpackaged(BaseMetadataApiCall):
    check_interval = 1
    soap_envelope_start = soap_envelopes.RETRIEVE_UNPACKAGED
    soap_envelope_status = soap_envelopes.CHECK_STATUS
    soap_envelope_result = soap_envelopes.CHECK_RETRIEVE_STATUS
    soap_action_start = "retrieve"
    soap_action_status = "checkStatus"
    soap_action_result = "checkRetrieveStatus"

    def __init__(self, task, package_xml, api_version):
        super(ApiRetrieveUnpackaged, self).__init__(task, api_version)
        self.package_xml = package_xml
        self._clean_package_xml()

    def _clean_package_xml(self):
        self.package_xml = re.sub(r"<\?xml.*\?>", "", self.package_xml)
        self.package_xml = re.sub("<Package.*>", "", self.package_xml, 1)
        self.package_xml = re.sub("</Package>", "", self.package_xml, 1)
        self.package_xml = re.sub("\n", "", self.package_xml)
        self.package_xml = re.sub(" +<", "<", self.package_xml)

    def _build_envelope_start(self):
        return self.soap_envelope_start.format(
            api_version=self.api_version, package_xml=self.package_xml
        )

    def _process_response(self, response):
        # Parse the metadata zip file from the response
        zipstr = parseString(response.content).getElementsByTagName("zipFile")
        if zipstr:
            zipstr = zipstr[0].firstChild.nodeValue
        zipstringio = io.BytesIO(base64.b64decode(zipstr))
        zipfile = ZipFile(zipstringio, "r")
        zipfile = zip_subfolder(zipfile, "unpackaged")
        return zipfile


class ApiRetrieveInstalledPackages(BaseMetadataApiCall):
    check_interval = 1
    soap_envelope_start = soap_envelopes.RETRIEVE_INSTALLEDPACKAGE
    soap_envelope_status = soap_envelopes.CHECK_STATUS
    soap_envelope_result = soap_envelopes.CHECK_RETRIEVE_STATUS
    soap_action_start = "retrieve"
    soap_action_status = "checkStatus"
    soap_action_result = "checkRetrieveStatus"

    def __init__(self, task, api_version=None):
        super(ApiRetrieveInstalledPackages, self).__init__(task, api_version)
        self.packages = {}

    def _process_response(self, response):
        # Parse the metadata zip file from the response
        zipstr = parseString(response.content).getElementsByTagName("zipFile")
        if zipstr:
            zipstr = zipstr[0].firstChild.nodeValue
        else:
            return self.packages
        zipstringio = io.BytesIO(base64.b64decode(zipstr))
        zipfile = ZipFile(zipstringio, "r")
        # Loop through all files in the zip skipping anything other than
        # InstalledPackages
        for path in zipfile.namelist():
            if not path.endswith(".installedPackage"):
                continue
            namespace = path.split("/")[-1].split(".")[0]
            version = parseString(zipfile.open(path).read()).getElementsByTagName(
                "versionNumber"
            )
            if version:
                version = version[0].firstChild.nodeValue
            self.packages[namespace] = version
        return self.packages


class ApiRetrievePackaged(BaseMetadataApiCall):
    check_interval = 1
    soap_envelope_start = soap_envelopes.RETRIEVE_PACKAGED
    soap_envelope_status = soap_envelopes.CHECK_STATUS
    soap_envelope_result = soap_envelopes.CHECK_RETRIEVE_STATUS
    soap_action_start = "retrieve"
    soap_action_status = "checkStatus"
    soap_action_result = "checkRetrieveStatus"

    def __init__(self, task, package_name, api_version):
        super(ApiRetrievePackaged, self).__init__(task, api_version)
        self.package_name = package_name

    def _build_envelope_start(self):
        return self.soap_envelope_start.format(
            api_version=self.api_version, package_name=escape(self.package_name)
        )

    def _process_response(self, response):
        # Parse the metadata zip file from the response
        zipstr = parseString(response.content).getElementsByTagName("zipFile")
        if zipstr:
            zipstr = zipstr[0].firstChild.nodeValue
        zipstringio = io.BytesIO(base64.b64decode(zipstr))
        zipfile = ZipFile(zipstringio, "r")
        return zipfile


class ApiDeploy(BaseMetadataApiCall):
    soap_envelope_start = soap_envelopes.DEPLOY
    soap_envelope_status = soap_envelopes.CHECK_DEPLOY_STATUS
    soap_action_start = "deploy"
    soap_action_status = "checkDeployStatus"

    def __init__(
        self,
        task,
        package_zip,
        purge_on_delete=None,
        api_version=None,
        check_only=False,
        test_level=None,
        run_tests=None,
    ):
        super(ApiDeploy, self).__init__(task, api_version)
        if purge_on_delete is None:
            purge_on_delete = True
        self._set_purge_on_delete(purge_on_delete)
        self.check_only = "true" if check_only else "false"
        self.test_level = test_level
        self.package_zip = package_zip
        self.run_tests = run_tests or []

    def _set_purge_on_delete(self, purge_on_delete):
        if not purge_on_delete or purge_on_delete == "false":
            self.purge_on_delete = "false"
        else:
            self.purge_on_delete = "true"
        # Disable purge on delete entirely for non sandbox or DE orgs as it is
        # not allowed
        org_type = self.task.org_config.org_type
        is_sandbox = self.task.org_config.is_sandbox
        if org_type != "Developer Edition" and not is_sandbox:
            self.purge_on_delete = "false"

    def _build_envelope_start(self):
        if self.package_zip:
            test_level = (
                f"<testLevel>{self.test_level}</testLevel>" if self.test_level else ""
            )
            run_tests = (
                "\n".join(f"<runTests>{x}</runTests>" for x in self.run_tests)
                if self.test_level == "RunSpecifiedTests"
                else ""
            )
            return self.soap_envelope_start.format(
                package_zip=self.package_zip,
                check_only=self.check_only,
                purge_on_delete=self.purge_on_delete,
                test_level=test_level,
                run_tests=run_tests,
                api_version=self.api_version,
            )

    def _process_response(self, response):
        resp_xml = parseString(response.content)
        status = resp_xml.getElementsByTagName("status")
        if status:
            status = status[0].firstChild.nodeValue
        else:
            # If no status element is in the result xml, return fail and log
            # the entire SOAP envelope in the log
            self._set_status("Failed", response.text)
            return self.status
        # Only done responses should be passed so we need to handle any status
        # related to done
        if status in ["Succeeded", "SucceededPartial"]:
            self._set_status("Success", status)
        else:
            # If failed, parse out the problem text and raise appropriate exception
            messages = []

            component_failures = resp_xml.getElementsByTagName("componentFailures")
            for component_failure in component_failures:
                problems = component_failure.getElementsByTagName("problem")
                problem_types = component_failure.getElementsByTagName("problemType")
                failure_info = {
                    "component_type": None,
                    "file_name": None,
                    "line_num": None,
                    "column_num": None,
                    "problem": problems[0].firstChild.nodeValue
                    if problems
                    else "Unknown problem",
                    "problem_type": problem_types[0].firstChild.nodeValue
                    if problem_types
                    else "Error",
                }
                failure_info["component_type"] = self._get_element_value(
                    component_failure, "componentType"
                )
                full_name = self._get_element_value(component_failure, "fullName")
                file_name = self._get_element_value(component_failure, "fileName")
                failure_info["file_name"] = full_name or file_name
                failure_info["line_num"] = self._get_element_value(
                    component_failure, "lineNumber"
                )
                failure_info["column_num"] = self._get_element_value(
                    component_failure, "columnNumber"
                )

                created = (
                    component_failure.getElementsByTagName("created")[
                        0
                    ].firstChild.nodeValue
                    == "true"
                )
                deleted = (
                    component_failure.getElementsByTagName("deleted")[
                        0
                    ].firstChild.nodeValue
                    == "true"
                )
                failure_info["action"] = self._get_action(created, deleted)

                if failure_info["file_name"] and failure_info["line_num"]:
                    messages.append(
                        "{action} of {component_type} {file_name}: {problem_type} on line {line_num}, col {column_num}: {problem}".format(
                            **failure_info
                        )
                    )
                elif failure_info["file_name"]:
                    messages.append(
                        "{action} of {component_type} {file_name}: {problem_type}: {problem}".format(
                            **failure_info
                        )
                    )
                else:
                    messages.append(
                        "{action} of {component_type}: {problem_type}: {problem}".format(
                            **failure_info
                        )
                    )

            if messages:
                # Deploy failures due to a component failure should raise MetadataComponentFailure
                log = "\n\n".join(messages)
                self._set_status("Failed", log)
                raise MetadataComponentFailure(log, response)

            else:
                problems = resp_xml.getElementsByTagName("problem")
                for problem in problems:
                    messages.append(problem.firstChild.nodeValue)
                errorMessages = resp_xml.getElementsByTagName("errorMessage")
                for errorMessage in errorMessages:
                    messages.append(errorMessage.firstChild.nodeValue)
                if messages:
                    log = "\n\n".join(messages)
                    raise MetadataApiError(log, response)

            # Parse out any failure text (from test failures in production
            # deployments) and add to log
            failures = resp_xml.getElementsByTagName("failures")
            for failure in failures:
                # Get needed values from subelements
                namespace = self._get_element_value(failure, "namespace")
                stacktrace = self._get_element_value(failure, "stackTrace")
                message = ["Apex Test Failure: "]
                if namespace:
                    message.append(f"from namespace {namespace}: ")
                if stacktrace:
                    message.append(stacktrace)
                messages.append("".join(message))

            if messages:
                # Deploy failures due to a component failure should raise MetadataComponentFailure
                log = "\n\n".join(messages)
                self._set_status("Failed", log)
                raise ApexTestException(log)
            else:
                log = response.text
                self._set_status("Failed", log)
                raise MetadataApiError(log, response)

        return self.status

    def _get_action(self, created, deleted):
        if created:
            return "Create"
        elif deleted:
            return "Delete"
        return "Update"


class ApiListMetadata(BaseMetadataApiCall):
    soap_envelope_start = soap_envelopes.LIST_METADATA
    soap_action_start = "listMetadata"

    def __init__(
        self, task, metadata_type, metadata=None, folder=None, as_of_version=None
    ):
        super(ApiListMetadata, self).__init__(task)
        self.metadata_type = metadata_type
        self.metadata = metadata
        self.folder = folder
        self.as_of_version = (
            as_of_version
            if as_of_version
            else task.project_config.project__package__api_version
        )
        self.api_version = self.as_of_version
        if self.metadata is None:
            self.metadata = defaultdict(list)

    def _build_envelope_start(self):
        folder = self.folder
        folder = f"\n      <folder>{folder}</folder>" if folder is not None else ""
        return self.soap_envelope_start.format(
            metadata_type=self.metadata_type,
            folder=folder,
            as_of_version=self.as_of_version,
        )

    def _process_response(self, response):
        metadata = []
        tags = [
            "createdById",
            "createdByName",
            "createdDate",
            "fileName",
            "fullName",
            "id",
            "lastModifiedById",
            "lastModifiedByName",
            "lastModifiedDate",
            "manageableState",
            "namespacePrefix",
            "type",
        ]
        # These tags will be interpreted into dates
        parse_dates = ["createdDate", "lastModifiedDate"]
        for result in parseString(response.content).getElementsByTagName("result"):
            result_data = {}
            # Parse fields
            for tag in tags:
                result_data[tag] = self._get_element_value(result, tag)
            # Parse dates
            for key in parse_dates:
                if result_data[key]:
                    try:
                        result_data[key] = parse_api_datetime(result_data[key])
                    except Exception as e:
                        raise MetadataParseError(
                            f"Could not parse a datetime in the MDAPI response: {str(e)}, {str(result)}",
                            response=response,
                        )
            metadata.append(result_data)
        self.metadata[self.metadata_type].extend(metadata)
        return self.metadata



from unittest import mock
import json
import os
import unittest

from cumulusci.core.config import OrgConfig
from cumulusci.tasks.salesforce.sourcetracking import ListChanges
from cumulusci.tasks.salesforce.sourcetracking import RetrieveChanges
from cumulusci.tasks.salesforce.sourcetracking import SnapshotChanges
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tests.util import create_project_config
from cumulusci.utils import temporary_dir


class TestListChanges(unittest.TestCase):
    """List the changes from a scratch org"""

    def test_run_task(self):
        task = create_task(ListChanges, {"exclude": "Ignore"})
        task._init_task()
        task.tooling = mock.Mock()
        task.logger = mock.Mock()
        task.tooling.query_all.return_value = {
            "totalSize": 1,
            "records": [
                {
                    "MemberType": "CustomObject",
                    "MemberName": "Test__c",
                    "RevisionNum": 1,
                },
                {
                    "MemberType": "CustomObject",
                    "MemberName": "Ignored__c",
                    "RevisionNum": 2,
                },
            ],
        }
        task._run_task()
        self.assertIn("CustomObject: Test__c", task.logger.info.call_args[0][0])

    def test_run_task__no_changes(self):
        task = create_task(ListChanges)
        task._init_task()
        task.tooling = mock.Mock()
        task.logger = mock.Mock()
        task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
        task._run_task()
        self.assertIn("Found no changes.", task.logger.info.call_args[0][0])

    def test_run_task__snapshot(self):
        with temporary_dir():
            task = create_task(ListChanges, {"snapshot": True})
            task._init_task()
            task.tooling = mock.Mock()
            messages = []
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionNum": 1,
                    }
                ],
            }
            task._run_task()
            self.assertTrue(
                os.path.exists(os.path.join(".cci", "snapshot", "test.json"))
            )
            self.assertIn("CustomObject: Test__c", messages)

            task = create_task(ListChanges)
            task._init_task()
            task.tooling = mock.Mock()
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionNum": 1,
                    }
                ],
            }
            task._run_task()
            self.assertIn("Found no changes.", messages)

    def test_filter_changes__include(self):
        foo = {"MemberType": "CustomObject", "MemberName": "foo__c", "RevisionNum": 1}
        bar = {"MemberType": "CustomObject", "MemberName": "bar__c", "RevisionNum": 1}
        foobar = {
            "MemberType": "CustomObject",
            "MemberName": "foobar__c",
            "RevisionNum": 1,
        }
        task = create_task(ListChanges, {"include": "foo", "exclude": "bar"})
        filtered, ignored = task._filter_changes([foo, bar, foobar])
        self.assertEqual([foo], filtered)

    def test_filter_changes__null_revnum(self):
        foo = {
            "MemberType": "CustomObject",
            "MemberName": "foo__c",
            "RevisionNum": None,
        }
        bar = {"MemberType": "CustomObject", "MemberName": "bar__c", "RevisionNum": 1}
        task = create_task(ListChanges, {})
        filtered, ignored = task._filter_changes([foo, bar])
        self.assertEqual([foo, bar], filtered)

    def test_load_maxrevision(self):
        with temporary_dir():
            task = create_task(ListChanges, {})
            task._store_maxrevision(1)
            assert task._load_maxrevision() == 1


@mock.patch("cumulusci.tasks.salesforce.sourcetracking.sfdx")
class TestRetrieveChanges(unittest.TestCase):
    """Retrieve changed components from a scratch org"""

    def test_init_options__sfdx_format(self, sfdx):
        with temporary_dir():
            project_config = create_project_config()
            project_config.project__source_format = "sfdx"
            with open("sfdx-project.json", "w") as f:
                json.dump(
                    {"packageDirectories": [{"path": "force-app", "default": True}]}, f
                )
            task = create_task(RetrieveChanges, {}, project_config)
            assert not task.md_format
            assert task.options["path"] == "force-app"

    def test_run_task(self, sfdx):
        sfdx_calls = []
        sfdx.side_effect = lambda cmd, *args, **kw: sfdx_calls.append(cmd)

        with temporary_dir():
            task = create_task(
                RetrieveChanges, {"include": "Test", "namespace_tokenize": "ns"}
            )
            task._init_task()
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionNum": 1,
                    }
                ],
            }

            task._run_task()

            assert sfdx_calls == [
                "force:mdapi:convert",
                "force:source:retrieve",
                "force:source:convert",
            ]
            assert os.path.exists(os.path.join("src", "package.xml"))

    def test_run_task__no_changes(self, sfdx):
        with temporary_dir() as path:
            task = create_task(RetrieveChanges, {"path": path})
            task._init_task()
            messages = []
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task._run_task()
            self.assertIn("No changes to retrieve", messages)


class TestSnapshotChanges(unittest.TestCase):
    def test_run_task(self):
        with temporary_dir():
            org_config = OrgConfig(
                {
                    "username": "test-cci@example.com",
                    "scratch": True,
                    "instance_url": "https://test.salesforce.com",
                    "access_token": "TOKEN",
                },
                "test",
            )
            task = create_task(SnapshotChanges, org_config=org_config)
            task._init_task()
            task.tooling.query = mock.Mock(
                side_effect=[
                    {"totalSize": 0, "records": [], "done": True},
                    {
                        "totalSize": 1,
                        "done": True,
                        "records": [
                            {
                                "MemberType": "CustomObject",
                                "MemberName": "Object2",
                                "RevisionNum": 1,
                            }
                        ],
                    },
                ]
            )
            task._run_task()
            self.assertTrue(
                os.path.exists(
                    os.path.join(
                        ".sfdx", "orgs", "test-cci@example.com", "maxrevision.json"
                    )
                )
            )

    def test_run_task__null_revnum(self):
        with temporary_dir():
            org_config = OrgConfig(
                {
                    "username": "test-cci@example.com",
                    "scratch": True,
                    "instance_url": "https://test.salesforce.com",
                    "access_token": "TOKEN",
                },
                "test",
            )
            task = create_task(SnapshotChanges, org_config=org_config)
            task._init_task()
            task.tooling.query = mock.Mock(
                return_value={
                    "totalSize": 1,
                    "done": True,
                    "records": [
                        {
                            "MemberType": "CustomObject",
                            "MemberName": "Object2",
                            "RevisionNum": None,
                        }
                    ],
                }
            )
            task._run_task()
            self.assertEqual(-1, task._snapshot["CustomObject"]["Object2"])
            self.assertFalse(
                os.path.exists(
                    os.path.join(
                        ".sfdx", "orgs", "test-cci@example.com", "maxrevision.json"
                    )
                )
            )

    def test_freeze(self):
        task = create_task(SnapshotChanges)
        steps = task.freeze(None)
        self.assertEqual([], steps)



import os.path

import unittest
from unittest import mock

from cumulusci.tasks.bulkdata import GenerateAndLoadData
from cumulusci.core.exceptions import TaskOptionsError

from .test_bulkdata import _make_task


class TestGenerateAndLoadData(unittest.TestCase):
    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_generate_and_load_data(self, _dataload):
        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        task = _make_task(
            GenerateAndLoadData,
            {
                "options": {
                    "num_records": 12,
                    "mapping": mapping_file,
                    "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                }
            },
        )

        task()
        _dataload.assert_called_once()

    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_generate_and_load_data_batched(self, _dataload):
        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        task = _make_task(
            GenerateAndLoadData,
            {
                "options": {
                    "num_records": 20,
                    "mapping": mapping_file,
                    "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                    "batch_size": 8,
                }
            },
        )

        task()
        calls = _dataload.mock_calls
        assert len(calls) == 3
        task_options = [call[1][0] for call in calls]  # get at the args
        for i in range(0, 3):
            assert task_options[i]["current_batch_number"] == i
            assert task_options[i]["batch_size"] == 8
            assert task_options[i]["mapping"] == mapping_file
        assert task_options[0]["num_records"] == task_options[1]["num_records"] == 8
        assert task_options[2]["num_records"] == 4

    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_batchsize_zero(self, _dataload):
        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        options = {
            "options": {
                "num_records": 20,
                "mapping": mapping_file,
                "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                "batch_size": 0,
            }
        }

        with self.assertRaises(TaskOptionsError):
            _make_task(GenerateAndLoadData, options)

    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_no_datageneration_task_specified(self, _dataload):
        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        options = {"options": {"num_records": 20, "mapping": mapping_file}}

        with self.assertRaises(TaskOptionsError) as e:
            _make_task(GenerateAndLoadData, options)

        assert "No data generation task" in str(e.exception)

    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_batchsize_matches_numrecords(self, _dataload):
        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        task = _make_task(
            GenerateAndLoadData,
            {
                "options": {
                    "num_records": 12,
                    "mapping": mapping_file,
                    "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                    "batch_size": 12,
                }
            },
        )

        task()
        _dataload.assert_called_once()

    def test_bad_mapping_file_path(self):
        with self.assertRaises(TaskOptionsError):
            _make_task(
                GenerateAndLoadData,
                {
                    "options": {
                        "num_records": 12,
                        "mapping": "does_not_exist",
                        "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                        "batch_size": 12,
                    }
                },
            )

    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_missing_mapping_file_when_needed(self, _dataload):
        with self.assertRaises(TaskOptionsError):
            task = _make_task(
                GenerateAndLoadData,
                {
                    "options": {
                        "num_records": 12,
                        "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                        "batch_size": 12,
                    }
                },
            )
            task()

    @mock.patch("cumulusci.tasks.bulkdata.GenerateAndLoadData._dataload")
    def test_missing_mapping_file_path(self, _dataload):
        task = _make_task(
            GenerateAndLoadData,
            {
                "options": {
                    "num_records": 12,
                    "data_generation_task": "cumulusci.tasks.bulkdata.tests.mock_data_factory_without_mapping.GenerateDummyData",
                    "batch_size": 12,
                }
            },
        )
        task()
        calls = _dataload.mock_calls
        assert calls[0][1][0]["generate_mapping_file"]

    def test_bad_options(self):
        with self.assertRaises(TaskOptionsError):
            _make_task(
                GenerateAndLoadData,
                {
                    "options": {
                        "num_records": 12,
                        "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                        "database_url": "not_a_real_url:///non-url",
                        "batch_size": -1,
                    }
                },
            )

    def test_loader_subtask(self):
        class MockLoadData:
            def __init__(self, *args, **kwargs):
                options = kwargs["task_config"].options
                assert options["num_records"] == 12
                assert options["database_url"].startswith("sqlite")
                assert options["bulk_mode"] == "Serial"
                assert "mapping_vanilla_sf" in options["mapping"]

            def __call__(self):
                pass

        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        with mock.patch(
            "cumulusci.tasks.bulkdata.generate_and_load_data.LoadData", MockLoadData
        ):
            task = _make_task(
                GenerateAndLoadData,
                {
                    "options": {
                        "num_records": 12,
                        "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                        "mapping": mapping_file,
                        "bulk_mode": "Serial",
                    }
                },
            )
            task()

    def test_error_on_tables_exist(self):
        class FakeEngine:
            pass

        class FakeMetadata:
            tables = {"foo": mock.MagicMock}

        def _setup_engine(*args, **kwargs):
            return FakeEngine(), FakeMetadata()

        with mock.patch(
            "cumulusci.tasks.bulkdata.generate_and_load_data.GenerateAndLoadData._setup_engine",
            _setup_engine,
        ):
            with self.assertRaises(TaskOptionsError):
                task = _make_task(
                    GenerateAndLoadData,
                    {
                        "options": {
                            "num_records": 12,
                            "data_generation_task": "cumulusci.tasks.bulkdata.tests.dummy_data_factory.GenerateDummyData",
                            "batch_size": 12,
                            "database_url": "not_a_real:///",
                        }
                    },
                )

                task()



import unittest
from unittest import mock

import cumulusci.robotframework.utils as robot_utils


class TestRobotframeworkUtils(unittest.TestCase):
    def setUp(self):
        robot_utils.BuiltIn = mock.Mock(name="BuiltIn")
        self.mock_selib = robot_utils.BuiltIn().get_library_instance("SeleniumLibrary")

    def test_screenshot_decorator_fail(self):
        """Verify that the decorator will capture a screenshot on keyword failure"""

        @robot_utils.capture_screenshot_on_error
        def example_function():
            raise Exception("Danger Will Robinson!")

        try:
            example_function()
        except Exception:
            pass
        self.mock_selib.failure_occurred.assert_called_once()

    def test_screenshot_decorator_pass(self):
        """Verify that decorator does NOT capture screenshot on keyword success"""

        @robot_utils.capture_screenshot_on_error
        def example_function():
            return True

        example_function()
        self.mock_selib.failure_occurred.assert_not_called()



import os
from typing import TextIO, Optional
from pathlib import Path
import shutil

import yaml


from cumulusci.core.utils import process_list_of_pairs_dict_arg

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.bulkdata.base_generate_data_task import BaseGenerateDataTask
from snowfakery.output_streams import SqlOutputStream
from snowfakery.data_generator import generate, StoppingCriteria
from snowfakery.generate_mapping_from_factory import mapping_from_factory_templates


class GenerateDataFromYaml(BaseGenerateDataTask):
    """Generate sample data from a YAML template file."""

    task_docs = """
    Depends on the currently un-released Snowfakery tool from SFDO. Better documentation
    will appear here when Snowfakery is publically available.
    """

    task_options = {
        **BaseGenerateDataTask.task_options,
        "generator_yaml": {
            "description": "A generator YAML file to use",
            "required": True,
        },
        "vars": {
            "description": "Pass values to override options in the format VAR1:foo,VAR2:bar"
        },
        "generate_mapping_file": {
            "description": "A path to put a mapping file inferred from the generator_yaml",
            "required": False,
        },
        "num_records_tablename": {
            "description": "A string representing which table to count records in."
        },
        "continuation_file": {
            "description": "YAML file generated by Snowfakery representing next steps for data generation"
        },
        "generate_continuation_file": {
            "description": "Path for Snowfakery to put its next continuation file"
        },
        "working_directory": {
            "description": "Default path for temporary / working files"
        },
    }
    stopping_criteria = None

    def __init__(self, *args, **kwargs):
        self.vars = {}
        super().__init__(*args, **kwargs)

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.yaml_file = os.path.abspath(self.options["generator_yaml"])
        if not os.path.exists(self.yaml_file):
            raise TaskOptionsError(f"Cannot find {self.yaml_file}")
        if "vars" in self.options:
            self.vars = process_list_of_pairs_dict_arg(self.options["vars"])
        self.generate_mapping_file = self.options.get("generate_mapping_file")
        if self.generate_mapping_file:
            self.generate_mapping_file = os.path.abspath(self.generate_mapping_file)
        num_records = self.options.get("num_records")
        if num_records is not None:
            num_records = int(num_records)
            num_records_tablename = self.options.get("num_records_tablename")

            if not num_records_tablename:
                raise TaskOptionsError(
                    f"Cannot specify num_records without num_records_tablename."
                )

            self.stopping_criteria = StoppingCriteria(
                num_records_tablename, num_records
            )
        self.working_directory = self.options.get("working_directory")

    def _generate_data(self, db_url, mapping_file_path, num_records, current_batch_num):
        """Generate all of the data"""
        if mapping_file_path:
            with open(mapping_file_path, "r") as f:
                self.mappings = yaml.safe_load(f)
        else:
            self.mappings = {}
        session, engine, base = self.init_db(db_url, self.mappings)
        self.logger.info(f"Generating batch {current_batch_num} with {num_records}")
        self.generate_data(session, engine, base, num_records, current_batch_num)
        session.commit()
        session.close()

    def default_continuation_file_path(self):
        return Path(self.working_directory) / "continuation.yml"

    def get_old_continuation_file(self) -> Optional[Path]:
        """Use a continuation file if specified or look for one in the working directory

        Return None if no file can be found.

        If this code is used within a GenerateAndLoad loop, the continuation files will go
        into the working directory specified by the GenerateAndLoad caller.
        """
        old_continuation_file = self.options.get("continuation_file")

        if old_continuation_file:
            old_continuation_file = Path(old_continuation_file)
            if not old_continuation_file.exists():
                raise TaskOptionsError(f"{old_continuation_file} does not exist")
        elif self.working_directory:
            path = self.default_continuation_file_path()
            if path.exists():
                old_continuation_file = path

        return old_continuation_file

    def open_new_continuation_file(self) -> Optional[TextIO]:
        """Create a continuation file based on config or working directory

        Return None if there is no config nor working directory.

        If this code is used within a GenerateAndLoad loop, the continuation files will go
        into the working directory specified by the GenerateAndLoad caller.
        """
        if self.options.get("generate_continuation_file"):
            new_continuation_file = open(
                self.options["generate_continuation_file"], "w+"
            )
        elif self.working_directory:
            new_continuation_file = open(
                Path(self.working_directory) / "continuation_next.yml", "w+"
            )
        else:
            new_continuation_file = None
        return new_continuation_file

    def generate_data(self, session, engine, base, num_records, current_batch_num):
        output_stream = SqlOutputStream.from_connection(
            session, engine, base, self.mappings
        )
        old_continuation_file = self.get_old_continuation_file()
        if old_continuation_file:
            # reopen to ensure file pointer is at starting point
            old_continuation_file = open(old_continuation_file, "r")
        new_continuation_file = self.open_new_continuation_file()

        with open(self.yaml_file) as open_yaml_file:
            summary = generate(
                open_yaml_file=open_yaml_file,
                user_options=self.vars,
                output_stream=output_stream,
                stopping_criteria=self.stopping_criteria,
                continuation_file=old_continuation_file,
                generate_continuation_file=new_continuation_file,
            )
            output_stream.close()

            if (
                new_continuation_file
                and Path(new_continuation_file.name).exists()
                and self.working_directory
            ):
                shutil.copyfile(
                    new_continuation_file.name, self.default_continuation_file_path()
                )

            if self.generate_mapping_file:
                with open(self.generate_mapping_file, "w+") as f:
                    yaml.safe_dump(
                        mapping_from_factory_templates(summary), f, sort_keys=False
                    )
                    f.close()



"""Wraps the github3 library to configure request retries."""

import os
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

import github3
from github3 import GitHub
from github3 import login
from github3.pulls import ShortPullRequest

from cumulusci.core.exceptions import GithubException


# Prepare request retry policy to be attached to github sessions.
# 401 is a weird status code to retry, but sometimes it happens spuriously
# and https://github.community/t5/GitHub-API-Development-and/Random-401-errors-after-using-freshly-generated-installation/m-p/22905 suggests retrying
retries = Retry(status_forcelist=(401, 502, 503, 504), backoff_factor=0.3)
adapter = HTTPAdapter(max_retries=retries)


def get_github_api(username=None, password=None):
    """Old API that only handles logging in as a user.

    Here for backwards-compatibility during the transition.
    """
    gh = login(username, password)
    gh.session.mount("http://", adapter)
    gh.session.mount("https://", adapter)
    return gh


INSTALLATIONS = {}


def get_github_api_for_repo(keychain, owner, repo):
    gh = GitHub()
    # Apply retry policy
    gh.session.mount("http://", adapter)
    gh.session.mount("https://", adapter)

    APP_KEY = os.environ.get("GITHUB_APP_KEY", "").encode("utf-8")
    APP_ID = os.environ.get("GITHUB_APP_ID")
    if APP_ID and APP_KEY:
        installation = INSTALLATIONS.get((owner, repo))
        if installation is None:
            gh.login_as_app(APP_KEY, APP_ID, expire_in=120)
            try:
                installation = gh.app_installation_for_repository(owner, repo)
            except github3.exceptions.NotFoundError:
                raise GithubException(
                    f"Could not access {owner}/{repo} using GitHub app. "
                    "Does the app need to be installed for this repository?"
                )
            INSTALLATIONS[(owner, repo)] = installation
        gh.login_as_app_installation(APP_KEY, APP_ID, installation.id)
    else:
        github_config = keychain.get_service("github")
        gh.login(github_config.username, github_config.password)
    return gh


def validate_service(options):
    username = options["username"]
    password = options["password"]
    gh = get_github_api(username, password)
    try:
        gh.rate_limit()
    except Exception as e:
        raise GithubException(f"Could not confirm access to the GitHub API: {str(e)}")


def get_pull_requests_with_base_branch(repo, base_branch_name, head=None, state=None):
    """Returns a list of pull requests with the given base branch"""
    if head:
        head = repo.owner.login + ":" + head
    return list(repo.pull_requests(base=base_branch_name, head=head, state=state))


def get_pull_requests_by_head(repo, branch_name):
    """Returns all pull requests with head equal to the given branch name."""
    if branch_name == repo.default_branch:
        return None

    return list(repo.pull_requests(head=repo.owner.login + ":" + branch_name))


def create_pull_request(repo, branch_name, base=None, title=None):
    """Creates a pull request for the given branch"""
    base = base or "master"
    title = title or "Auto-Generated Pull Request"
    pull_request = repo.create_pull(title, base, branch_name)
    return pull_request


def add_labels_to_pull_request(repo, pull_request, *labels):
    """Adds a label to a pull request via the issue object
    Args:
    * repo: Repository object
    * pull_request: ShortPullRequest object that exists in repo
    * labels: list(str) of labels to add to the pull request"""
    issue = repo.issue(pull_request.number)
    issue.add_labels(*labels)


def is_label_on_pull_request(repo, pull_request, label_name):
    """Returns True if the given label is on the pull request with the given
    pull request number. False otherwise."""
    labels = list(repo.issue(pull_request.number).labels())
    return any(label_name == issue_label.name for issue_label in labels)


def get_pull_requests_by_commit(github, repo, commit_sha):
    endpoint = (
        github.session.base_url
        + f"/repos/{repo.owner.login}/{repo.name}/commits/{commit_sha}/pulls"
    )
    response = github.session.get(
        endpoint, headers={"Accept": "application/vnd.github.groot-preview+json"}
    )
    json_list = response.json()

    # raises github3.exceptions.IncompleteResposne
    # when these are not present
    for json in json_list:
        json["body_html"] = ""
        json["body_text"] = ""

    return [ShortPullRequest(json, github) for json in json_list]


def is_pull_request_merged(pull_request):
    """Takes a github3.pulls.ShortPullRequest object"""
    return pull_request.merged_at is not None


def markdown_link_to_pr(change_note):
    return f"{change_note.title} [[PR{change_note.number}]({change_note.html_url})]"


def find_latest_release(repo, include_beta=None):
    try:
        if include_beta:
            return next(repo.releases())
        else:
            return repo.latest_release()
    except (github3.exceptions.NotFoundError, StopIteration):
        pass


def find_previous_release(repo, prefix=None):
    most_recent = None
    for release in repo.releases():
        if prefix and not release.tag_name.startswith(prefix):
            continue
        if not prefix and release.prerelease:
            continue
        # Return the second release
        if most_recent is None:
            most_recent = release
        else:
            return release


def create_gist(github, description, files):
    """Creates a gist with the given description and files.

    github - an
    description - str
    files - A dict of files in the form of {filename:{'content': content},...}
    """
    return github.create_gist(description, files, public=False)



from lxml import etree
import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.metadata_etl import AddPermissionSetPermissions
from cumulusci.tasks.metadata_etl import MD

PERMSET_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<PermissionSet xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>CustomApp</application>
        <visible>false</visible>
    </applicationVisibilities>
    <classAccesses>
        <apexClass>ApexController</apexClass>
        <enabled>false</enabled>
    </classAccesses>
    <fieldPermissions>
        <editable>false</editable>
        <field>Test__c.Lookup__c</field>
        <readable>false</readable>
    </fieldPermissions>
    <hasActivationRequired>false</hasActivationRequired>
    <label>Test</label>
    <objectPermissions>
        <allowCreate>false</allowCreate>
        <allowDelete>false</allowDelete>
        <allowEdit>false</allowEdit>
        <allowRead>false</allowRead>
        <modifyAllRecords>false</modifyAllRecords>
        <object>Test__c</object>
        <viewAllRecords>false</viewAllRecords>
    </objectPermissions>
    <recordTypeVisibilities>
        <recordType>Case.Test</recordType>
        <visible>true</visible>
    </recordTypeVisibilities>
    <tabSettings>
        <tab>standard-report</tab>
        <visibility>Visible</visibility>
    </tabSettings>
    <userPermissions>
        <enabled>true</enabled>
        <name>ActivitiesAccess</name>
    </userPermissions>
</PermissionSet>
"""


class TestAddPermissionSetPermissions:
    def test_adds_new_field_permission(self):
        task = create_task(
            AddPermissionSetPermissions,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "field_permissions": [
                    {
                        "field": "Test__c.Description__c",
                        "readable": True,
                        "editable": True,
                    }
                ],
            },
        )

        tree = etree.fromstring(PERMSET_XML).getroottree()

        assert (
            len(
                tree.findall(
                    f".//{MD}fieldPermissions[{MD}field='Test__c.Description__c']"
                )
            )
            == 0
        )

        result = task._transform_entity(tree, "PermSet")

        fieldPermissions = result.findall(
            f".//{MD}fieldPermissions[{MD}field='Test__c.Description__c']"
        )
        assert len(fieldPermissions) == 1
        readable = fieldPermissions[0].findall(f".//{MD}readable")
        assert len(readable) == 1
        assert readable[0].text == "true"
        editable = fieldPermissions[0].findall(f".//{MD}editable")
        assert len(editable) == 1
        assert editable[0].text == "true"

    def test_updates_existing_field_permission(self):
        task = create_task(
            AddPermissionSetPermissions,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "field_permissions": [
                    {"field": "Test__c.Lookup__c", "readable": True, "editable": True}
                ],
            },
        )

        tree = etree.fromstring(PERMSET_XML).getroottree()

        assert (
            len(tree.findall(f".//{MD}fieldPermissions[{MD}field='Test__c.Lookup__c']"))
            == 1
        )

        result = task._transform_entity(tree, "PermSet")

        fieldPermissions = result.findall(
            f".//{MD}fieldPermissions[{MD}field='Test__c.Lookup__c']"
        )
        assert len(fieldPermissions) == 1
        readable = fieldPermissions[0].findall(f".//{MD}readable")
        assert len(readable) == 1
        assert readable[0].text == "true"
        editable = fieldPermissions[0].findall(f".//{MD}editable")
        assert len(editable) == 1
        assert editable[0].text == "true"

    def test_adds_new_class_permission(self):
        task = create_task(
            AddPermissionSetPermissions,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "class_accesses": [{"apexClass": "LWCController", "enabled": True}],
            },
        )

        tree = etree.fromstring(PERMSET_XML).getroottree()

        assert (
            len(tree.findall(f".//{MD}classAccesses[{MD}apexClass='LWCController']"))
            == 0
        )

        result = task._transform_entity(tree, "PermSet")

        classAccesses = result.findall(
            f".//{MD}classAccesses[{MD}apexClass='LWCController']"
        )
        assert len(classAccesses) == 1
        enabled = classAccesses[0].findall(f".//{MD}enabled")
        assert len(enabled) == 1
        assert enabled[0].text == "true"

    def test_upserts_existing_class_permission(self):
        task = create_task(
            AddPermissionSetPermissions,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "class_accesses": [{"apexClass": "ApexController", "enabled": True}],
            },
        )

        tree = etree.fromstring(PERMSET_XML).getroottree()

        assert (
            len(tree.findall(f".//{MD}classAccesses[{MD}apexClass='ApexController']"))
            == 1
        )

        result = task._transform_entity(tree, "PermSet")

        classAccesses = result.findall(
            f".//{MD}classAccesses[{MD}apexClass='ApexController']"
        )
        assert len(classAccesses) == 1
        enabled = classAccesses[0].findall(f".//{MD}enabled")
        assert len(enabled) == 1
        assert enabled[0].text == "true"

    def test_missing_apexclass_throws_exception(self):
        task = create_task(
            AddPermissionSetPermissions,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "class_accesses": [{"enabled": True}],
            },
        )

        tree = etree.fromstring(PERMSET_XML).getroottree()

        with pytest.raises(TaskOptionsError):
            task._transform_entity(tree, "PermSet")

    def test_missing_field_throws_exception(self):
        task = create_task(
            AddPermissionSetPermissions,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "field_permissions": [{"readable": True}],
            },
        )

        tree = etree.fromstring(PERMSET_XML).getroottree()

        with pytest.raises(TaskOptionsError):
            task._transform_entity(tree, "PermSet")



from lxml import etree

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata_etl import (
    MetadataSingleEntityTransformTask,
    get_new_tag_index,
    MD,
)


class AddValueSetEntries(MetadataSingleEntityTransformTask):
    entity = "StandardValueSet"
    task_options = {
        "entries": {
            "description": "Array of standardValues to insert. "
            "Each standardValue should contain the keys 'fullName', the API name of the entry, "
            "and 'label', the user-facing label. OpportunityStage entries require the additional "
            "keys 'closed', 'won', 'forecastCategory', and 'probability'; CaseStatus entries "
            "require 'closed'.",
            "required": False,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _transform_entity(self, metadata, api_name):
        for entry in self.options.get("entries", []):
            if "fullName" not in entry or "label" not in entry:
                raise TaskOptionsError(
                    "Standard value set entries must contain the 'fullName' and 'label' keys."
                )

            # Check for extra metadata on CaseStatus and OpportunityStage
            if api_name == "OpportunityStage":
                if not all(
                    [
                        "closed" in entry,
                        "forecastCategory" in entry,
                        "probability" in entry,
                        "won" in entry,
                    ]
                ):
                    raise TaskOptionsError(
                        "OpportunityStage standard value set entries require the keys "
                        "'closed', 'forecastCategory', 'probability', and 'won'"
                    )
            if api_name == "CaseStatus":
                if "closed" not in entry:
                    raise TaskOptionsError(
                        "CaseStatus standard value set entries require the key 'closed'"
                    )

            new_entry_index = get_new_tag_index(metadata, "standardValue")

            existing_entry = metadata.findall(
                f".//{MD}standardValue[{MD}fullName='{entry['fullName']}']"
            )
            if not existing_entry:
                # Entry doesn't exist. Insert it.
                elem = etree.Element(f"{MD}standardValue")
                metadata.getroot().insert(new_entry_index, elem)

                elem_fullName = etree.SubElement(elem, f"{MD}fullName")
                elem_fullName.text = entry["fullName"]

                elem_label = etree.SubElement(elem, f"{MD}label")
                elem_label.text = entry["label"]

                elem_default = etree.SubElement(elem, f"{MD}default")
                elem_default.text = "false"

                if api_name in ["OpportunityStage", "CaseStatus"]:
                    elem_closed = etree.SubElement(elem, f"{MD}closed")
                    elem_closed.text = str(entry["closed"]).lower()

                if api_name == "OpportunityStage":
                    elem_won = etree.SubElement(elem, f"{MD}won")
                    elem_won.text = str(entry["won"]).lower()

                    elem_probability = etree.SubElement(elem, f"{MD}probability")
                    elem_probability.text = str(entry["probability"])

                    elem_forecast_category = etree.SubElement(
                        elem, f"{MD}forecastCategory"
                    )
                    elem_forecast_category.text = entry["forecastCategory"]

        return metadata
